import unittest

import pyglet
import pyglet.text.layout
from pyglet.text import runlist
from pyglet.text.formats.attributed import AttributedTextDecoder


class TestStyleRuns(unittest.TestCase):

    def check_value(self, runs, value):
        for i, style in enumerate(value):
            self.assertTrue(runs[i] == style, repr(runs.runs))
        self.check_optimal(runs)
        self.check_continuous(runs)
        self.check_iter(runs, value)
        self.check_iter_range(runs, value)

    def check_optimal(self, runs):
        last_style = None
        for _, _, style in runs:
            self.assertTrue(style != last_style)
            last_style = style

    def check_continuous(self, runs):
        next_start = 0
        for start, end, _ in runs:
            self.assertTrue(start == next_start)
            next_start = end

    def check_iter(self, runs, value):
        for start, end, style in runs:
            for v in value[start:end]:
                self.assertTrue(v == style)

    def check_iter_range(self, runs, value):
        for interval in range(1, len(value)):
            it = runs.get_run_iterator()
            for start in range(0, len(value), interval):
                end = min(start + interval, len(value))
                for s, e, style in it.ranges(start, end):
                    for v in value[s:e]:
                        self.assertTrue(v == style, (start, end, s, e, style))

    def check_empty(self, runs, value):
        start, end, style = next(iter(runs))
        self.assertTrue(start == 0)
        self.assertTrue(end == 0)
        self.assertTrue(style == value)

    def test_zero(self):
        runs = runlist.RunList(0, 'x')
        it = iter(runs)

        start, end, s = next(it)
        self.assertTrue(start == 0)
        self.assertTrue(end == 0)
        self.assertTrue(s == 'x')

        with self.assertRaises(StopIteration):
            next(it)
        self.check_optimal(runs)

    def test_initial(self):
        runs = runlist.RunList(10, 'x')
        it = iter(runs)

        start, end, s = next(it)
        self.assertTrue(start == 0)
        self.assertTrue(end == 10)
        self.assertTrue(s == 'x')

        with self.assertRaises(StopIteration):
            next(it)

        self.check_value(runs, 'x' * 10)

    def test_set1(self):
        runs = runlist.RunList(10, 'a')
        runs.set_run(2, 8, 'b')
        self.check_value(runs, 'aabbbbbbaa')

    def test_set1_start(self):
        runs = runlist.RunList(10, 'a')
        runs.set_run(0, 5, 'b')
        self.check_value(runs, 'bbbbbaaaaa')

    def test_set1_end(self):
        runs = runlist.RunList(10, 'a')
        runs.set_run(5, 10, 'b')
        self.check_value(runs, 'aaaaabbbbb')

    def test_set1_all(self):
        runs = runlist.RunList(10, 'a')
        runs.set_run(0, 10, 'b')
        self.check_value(runs, 'bbbbbbbbbb')

    def test_set1_1(self):
        runs = runlist.RunList(10, 'a')
        runs.set_run(1, 2, 'b')
        self.check_value(runs, 'abaaaaaaaa')

    def test_set_overlapped(self):
        runs = runlist.RunList(10, 'a')
        runs.set_run(0, 5, 'b')
        self.check_value(runs, 'bbbbbaaaaa')
        runs.set_run(5, 10, 'c')
        self.check_value(runs, 'bbbbbccccc')
        runs.set_run(3, 7, 'd')
        self.check_value(runs, 'bbbddddccc')
        runs.set_run(4, 6, 'e')
        self.check_value(runs, 'bbbdeedccc')
        runs.set_run(5, 9, 'f')
        self.check_value(runs, 'bbbdeffffc')
        runs.set_run(2, 3, 'g')
        self.check_value(runs, 'bbgdeffffc')
        runs.set_run(1, 3, 'h')
        self.check_value(runs, 'bhhdeffffc')
        runs.set_run(1, 9, 'i')
        self.check_value(runs, 'biiiiiiiic')
        runs.set_run(0, 10, 'j')
        self.check_value(runs, 'jjjjjjjjjj')

    def test_insert_empty(self):
        runs = runlist.RunList(0, 'a')
        runs.insert(0, 10)
        self.check_value(runs, 'aaaaaaaaaa')

    def test_insert_beginning(self):
        runs = runlist.RunList(5, 'a')
        runs.set_run(1, 4, 'b')
        self.check_value(runs, 'abbba')
        runs.insert(0, 3)
        self.check_value(runs, 'aaaabbba')

    def test_insert_beginning_1(self):
        runs = runlist.RunList(5, 'a')
        self.check_value(runs, 'aaaaa')
        runs.insert(0, 1)
        runs.set_run(0, 1, 'a')
        self.check_value(runs, 'aaaaaa')
        runs.insert(0, 1)
        runs.set_run(0, 1, 'a')
        self.check_value(runs, 'aaaaaaa')
        runs.insert(0, 1)
        runs.set_run(0, 1, 'a')
        self.check_value(runs, 'aaaaaaaa')

    def test_insert_beginning_2(self):
        runs = runlist.RunList(5, 'a')
        self.check_value(runs, 'aaaaa')
        runs.insert(0, 1)
        runs.set_run(0, 1, 'b')
        self.check_value(runs, 'baaaaa')
        runs.insert(0, 1)
        runs.set_run(0, 1, 'c')
        self.check_value(runs, 'cbaaaaa')
        runs.insert(0, 1)
        runs.set_run(0, 1, 'c')
        self.check_value(runs, 'ccbaaaaa')

    def test_insert_1(self):
        runs = runlist.RunList(5, 'a')
        runs.set_run(1, 4, 'b')
        self.check_value(runs, 'abbba')
        runs.insert(1, 3)
        self.check_value(runs, 'aaaabbba')

    def test_insert_2(self):
        runs = runlist.RunList(5, 'a')
        runs.set_run(1, 2, 'b')
        self.check_value(runs, 'abaaa')
        runs.insert(2, 3)
        self.check_value(runs, 'abbbbaaa')

    def test_insert_end(self):
        runs = runlist.RunList(5, 'a')
        runs.set_run(4, 5, 'b')
        self.check_value(runs, 'aaaab')
        runs.insert(5, 3)
        self.check_value(runs, 'aaaabbbb')

    def test_insert_almost_end(self):
        runs = runlist.RunList(5, 'a')
        runs.set_run(0, 3, 'b')
        runs.set_run(4, 5, 'c')
        self.check_value(runs, 'bbbac')
        runs.insert(4, 3)
        self.check_value(runs, 'bbbaaaac')

    def test_delete_1_beginning(self):
        runs = runlist.RunList(5, 'a')
        self.check_value(runs, 'aaaaa')
        runs.delete(0, 3)
        self.check_value(runs, 'aa')

    def test_delete_1_middle(self):
        runs = runlist.RunList(5, 'a')
        self.check_value(runs, 'aaaaa')
        runs.delete(1, 4)
        self.check_value(runs, 'aa')

    def test_delete_1_end(self):
        runs = runlist.RunList(5, 'a')
        self.check_value(runs, 'aaaaa')
        runs.delete(2, 5)
        self.check_value(runs, 'aa')

    def test_delete_1_all(self):
        runs = runlist.RunList(5, 'a')
        self.check_value(runs, 'aaaaa')
        runs.delete(0, 5)
        self.check_value(runs, '')
        self.check_empty(runs, 'a')

    def create_runs1(self):
        runs = runlist.RunList(10, 'a')
        runs.set_run(1, 10, 'b')
        runs.set_run(2, 10, 'c')
        runs.set_run(3, 10, 'd')
        runs.set_run(4, 10, 'e')
        runs.set_run(5, 10, 'f')
        runs.set_run(6, 10, 'g')
        runs.set_run(7, 10, 'h')
        runs.set_run(8, 10, 'i')
        runs.set_run(9, 10, 'j')
        self.check_value(runs, 'abcdefghij')
        return runs

    def create_runs2(self):
        runs = runlist.RunList(10, 'a')
        runs.set_run(4, 7, 'b')
        runs.set_run(7, 10, 'c')
        self.check_value(runs, 'aaaabbbccc')
        return runs

    def test_delete2(self):
        runs = self.create_runs1()
        runs.delete(0, 5)
        self.check_value(runs, 'fghij')

    def test_delete3(self):
        runs = self.create_runs1()
        runs.delete(2, 8)
        self.check_value(runs, 'abij')

    def test_delete4(self):
        runs = self.create_runs2()
        runs.delete(0, 5)
        self.check_value(runs, 'bbccc')

    def test_delete5(self):
        runs = self.create_runs2()
        runs.delete(5, 10)
        self.check_value(runs, 'aaaab')

    def test_delete6(self):
        runs = self.create_runs2()
        runs.delete(0, 8)
        self.check_value(runs, 'cc')

    def test_delete7(self):
        runs = self.create_runs2()
        runs.delete(2, 10)
        self.check_value(runs, 'aa')

    def test_delete8(self):
        runs = self.create_runs2()
        runs.delete(3, 8)
        self.check_value(runs, 'aaacc')

    def test_delete9(self):
        runs = self.create_runs2()
        runs.delete(7, 8)
        self.check_value(runs, 'aaaabbbcc')

    def test_delete10(self):
        runs = self.create_runs2()
        runs.delete(8, 9)
        self.check_value(runs, 'aaaabbbcc')

    def test_delete11(self):
        runs = self.create_runs2()
        runs.delete(9, 10)
        self.check_value(runs, 'aaaabbbcc')

    def test_delete12(self):
        runs = self.create_runs2()
        runs.delete(4, 5)
        self.check_value(runs, 'aaaabbccc')

    def test_delete13(self):
        runs = self.create_runs2()
        runs.delete(5, 6)
        self.check_value(runs, 'aaaabbccc')

    def test_delete14(self):
        runs = self.create_runs2()
        runs.delete(6, 7)
        self.check_value(runs, 'aaaabbccc')


class TestIssues(unittest.TestCase):

    def test_issue471(self):
        doc = pyglet.text.document.FormattedDocument()
        layout = pyglet.text.layout.IncrementalTextLayout(doc, 100, 100, width=500, height=100)
        doc.insert_text(0, "hello", {'bold': True})
        doc.text = ""

    def test_issue471_comment2(self):
        doc2 = pyglet.text.decode_attributed('{bold True}a')
        layout = pyglet.text.layout.IncrementalTextLayout(doc2, 100, 10, width=500, height=100)
        layout.document.delete_text(0, len(layout.document.text))

    def test_issue241_comment4a(self):
        document = pyglet.text.document.FormattedDocument("")
        layout = pyglet.text.layout.IncrementalTextLayout(document, 50, 50, width=500, height=100)
        document.set_style(0, len(document.text), {"font_name": "Arial"})

    def test_issue241_comment4b(self):
        document = pyglet.text.document.FormattedDocument("test")
        layout = pyglet.text.layout.IncrementalTextLayout(document, 50, 50, width=500, height=100)
        document.set_style(0, len(document.text), {"font_name": "Arial"})
        document.delete_text(0, len(document.text))

    def test_issue241_comment5(self):
        document = pyglet.text.document.FormattedDocument('A')
        document.set_style(0, 1, dict(bold=True))
        layout = pyglet.text.layout.IncrementalTextLayout(document, 100, 100, width=500, height=100)
        document.delete_text(0, 1)

    def test_issue429_comment4a(self):
        doc = pyglet.text.decode_attributed(
            '{bold True}Hello{bold False}\n\n\n\n')
        doc2 = pyglet.text.decode_attributed(
            '{bold True}Goodbye{bold False}\n\n\n\n')
        layout = pyglet.text.layout.IncrementalTextLayout(doc, 100, 10, width=500, height=100)
        layout.document = doc2
        layout.document.delete_text(0, len(layout.document.text))

    def test_issue429_comment4b(self):
        doc2 = pyglet.text.decode_attributed('{bold True}a{bold False}b')
        layout = pyglet.text.layout.IncrementalTextLayout(doc2, 100, 10, width=500, height=100)
        layout.document.delete_text(0, len(layout.document.text))


class AttributedTextDecoderTests(unittest.TestCase):

    def testOneNewlineBecomesSpace(self):
        doc = AttributedTextDecoder().decode('one\ntwo')
        self.assertEqual('one two', doc.text)

    def testTwoNewlinesBecomesParagraph(self):
        from pyglet.text.formats.attributed import AttributedTextDecoder

        doc = AttributedTextDecoder().decode('one\n\ntwo')
        self.assertEqual('one\ntwo', doc.text)



"""Bindings, constants, and helpers for Mac's CoreAudio

Relevant Apple documentation is located at:

* https://developer.apple.com/documentation/audiotoolbox/
* https://developer.apple.com/documentation/coreaudio

This module includes:

* ctypes structs and function bindings
* Constants for audio file read and decoding
* Error checking helpers

"""
from ctypes import c_void_p, c_int, c_bool, Structure, c_uint32, util, cdll, c_uint, c_double, POINTER, c_int64, \
    CFUNCTYPE
from typing import Final

from pyglet.libs.darwin import CFURLRef

lib = util.find_library('CoreAudio')

if lib is None:
    lib = '/System/Library/Frameworks/CoreAudio.framework/CoreAudio'

ca = cdll.LoadLibrary(lib)


class AudioStreamPacketDescription(Structure):
    _fields_ = [
        ('mStartOffset', c_int64),
        ('mVariableFramesInPacket', c_uint32),
        ('mDataByteSize', c_uint32)
    ]


class AudioStreamBasicDescription(Structure):
    _fields_ = [
        ('mSampleRate', c_double),
        ('mFormatID', c_uint32),
        ('mFormatFlags', c_uint32),
        ('mBytesPerPacket', c_uint32),
        ('mFramesPerPacket', c_uint32),
        ('mBytesPerFrame', c_uint32),
        ('mChannelsPerFrame', c_uint32),
        ('mBitsPerChannel', c_uint32),
        ('mReserved', c_uint32)
    ]

    def __repr__(self):
        return f"AudioStreamBasicDescription(sample_rate={self.mSampleRate}, channels={self.mChannelsPerFrame}, " \
               f"fmt={self.mFormatID}, bytes_per_packet={self.mBytesPerPacket}, bits={self.mBitsPerChannel}, " \
               f"frames_per_packet={self.mFramesPerPacket}, bytes_per_frame={self.mBytesPerFrame})"


class AudioBuffer(Structure):
    _fields_ = [
        ("mNumberChannels", c_uint),
        ("mDataByteSize", c_uint),
        ("mData", c_void_p),
    ]


class AudioBufferList(Structure):
    _fields_ = [
        ("mNumberBuffers", c_uint),
        ("mBuffers", AudioBuffer * 1),
    ]


kCFURLPOSIXPathStyle = 0
ExtAudioFilePropertyID = c_uint32
OSStatus = c_int

ExtAudioFileRef = c_void_p

ca.ExtAudioFileOpenURL.restype = c_int
ca.ExtAudioFileOpenURL.argtypes = [CFURLRef, ExtAudioFileRef]

ca.ExtAudioFileGetProperty.restype = c_int
ca.ExtAudioFileGetProperty.argtypes = [ExtAudioFileRef, ExtAudioFilePropertyID, POINTER(c_uint32), c_void_p]

ca.ExtAudioFileSetProperty.restype = c_int
ca.ExtAudioFileSetProperty.argtypes = [ExtAudioFileRef, ExtAudioFilePropertyID, c_uint32, c_void_p]

ca.ExtAudioFileOpenURL.restype = OSStatus
ca.ExtAudioFileOpenURL.argtypes = [CFURLRef, ExtAudioFileRef]

AudioFileTypeID = c_uint32
AudioFileID = c_void_p

AudioFile_ReadProc = CFUNCTYPE(c_int, c_void_p, c_int64, c_uint32, c_void_p, POINTER(c_uint32))
AudioFile_GetSizeProc = CFUNCTYPE(c_int64, c_void_p)

ca.AudioFileOpenWithCallbacks.restype = OSStatus
ca.AudioFileOpenWithCallbacks.argtypes = [c_void_p, AudioFile_ReadProc, c_void_p, AudioFile_GetSizeProc, c_void_p,
                                          AudioFileTypeID, POINTER(AudioFileID)]

ca.ExtAudioFileWrapAudioFileID.restype = OSStatus
ca.ExtAudioFileWrapAudioFileID.argtypes = [AudioFileID, c_bool, POINTER(ExtAudioFileRef)]

ca.ExtAudioFileRead.restype = OSStatus
ca.ExtAudioFileRead.argtypes = [ExtAudioFileRef, POINTER(c_uint32), POINTER(AudioBufferList)]

ca.ExtAudioFileSeek.restype = OSStatus
ca.ExtAudioFileSeek.argtypes = [ExtAudioFileRef, c_int64]

ca.ExtAudioFileDispose.restype = OSStatus
ca.ExtAudioFileDispose.argtypes = [ExtAudioFileRef]

ca.AudioFileClose.restype = OSStatus
ca.AudioFileClose.argtypes = [AudioFileID]


def c_literal(mnemonic: str) -> int:
    """Pack a tiny ASCII string into a 32-bit int.

    Example: 'xyz' -> 0x78797a (Base 10: 7895418)

    Although many CoreAudio constants use this function, the only
    consistent rule seems to be a max of 4 ASCII characters to fit
    into a 32-bit int.

    Otherwise, error code constants may follow a loose convention:

    * '?' at the start error codes for unsupported actions
      Example: c_literal('?wht') # as in "what?"

    * '!' at the end of error codes for unintelligible data
      Example: c_literal('?siz') # "this data is the wrong size!"

    Args:
        mnemonic:
            Up to 4 ASCII characters to shift left
    Returns:
        A 32-bit int equivalent of the string.
    """
    num = 0
    for idx, char in enumerate(mnemonic):
        num |= ord(char) << (len(mnemonic) - idx - 1) * 8
    return num


# Non-error file & format constants
kAudioFilePropertyMagicCookieData = c_literal('mgic')
kExtAudioFileProperty_FileDataFormat = c_literal('ffmt')
kExtAudioFileProperty_ClientDataFormat = c_literal('cfmt')
kExtAudioFileProperty_FileLengthFrames = c_literal('#frm')
kAudioFormatLinearPCM = c_literal('lpcm')

kAudioFormatFlagIsFloat = 1 << 0
kAudioFormatFlagIsBigEndian = 1 << 1
kAudioFormatFlagIsSignedInteger = 1 << 2
kAudioFormatFlagIsPacked = 1 << 3
kAudioFormatFlagsNativeEndian = 0
kAudioFormatFlagsCanonical = kAudioFormatFlagIsFloat | kAudioFormatFlagsNativeEndian | kAudioFormatFlagIsPacked
kAudioQueueProperty_MagicCookie = c_literal('aqmc')


# ERRORS:

# General System errors
kAudio_UnimplementedError = -4
kAudio_FileNotFoundError = -43
kAudio_ParamError = -50
kAudio_MemFullError = -108


# All error constants below correspond to identically named errors in
# Apple's audiotoolbox. The doc for each is at URLs ending with the same
# names. For example, kAudioFileUnspecifiedError's documentation is at:
# https://developer.apple.com/documentation/audiotoolbox/

# General file read errors
kAudioFileNotOpenError = -38
kAudioFileEndOfFileError = -39
kAudioFilePositionError = -40
kAudioFileFileNotFoundError = -43

# File access mnemonic codes                    # Hex       , Base 10
kAudioFileUnspecifiedError = c_literal('wht?')  # 0x7768743F, 2003334207
kAudioFileUnsupportedFileTypeError = c_literal('typ?')  # 0x7479703F, 1954115647
kAudioFileUnsupportedDataFormatError = c_literal('fmt?')  # 0x666D743F, 1718449215
kAudioFileUnsupportedPropertyError = c_literal('pty?')  # 0x7074793F, 1886681407
kAudioFileBadPropertySizeError = c_literal('!siz')  # 0x2173697A,  561211770
kAudioFilePermissionsError = c_literal('prm?')  # 0x70726D3F, 1886547263
kAudioFileNotOptimizedError = c_literal('optm')  # 0x6F70746D, 1869640813

# Format-specific error codes                    # Hex       , Base 10
kAudioFileInvalidChunkError = c_literal('chk?')  # 0x63686B3F, 1667787583
kAudioFileDoesNotAllow64BitDataSizeError = c_literal('off?')  # 0x6F66663F, 1868981823
kAudioFileInvalidPacketOffsetError = c_literal('pck?')  # 0x70636B3F, 1885563711
kAudioFileInvalidFileError = c_literal('dta?')  # 0x6474613F, 1685348671
kAudioFileOperationNotSupportedError = c_literal('op?')  # 0x6F703F3F, 1869627199


# Maps kAudio errors -> error text
err_str_db: Final[dict[int, str]] = {
    kAudioFileNotOpenError: "The file is closed.",
    kAudioFileEndOfFileError: "End of file.",
    kAudioFilePositionError: "Invalid file position.",
    kAudioFileFileNotFoundError: "File not found.",
    kAudioFileUnspecifiedError: "An unspecified error has occurred.",
    kAudioFileUnsupportedFileTypeError: "The file type is not supported.",
    kAudioFileUnsupportedDataFormatError: "The data format is not supported by this file type.",
    kAudioFileUnsupportedPropertyError: "The property is not supported.",
    kAudioFileBadPropertySizeError: "The size of the property data was not correct.",
    kAudioFilePermissionsError: "The operation violated the file permissions.",
    kAudioFileNotOptimizedError: "The chunks following the audio data chunk are preventing the extension of the audio data chunk. To write more data, you must optimize the file.",
    kAudioFileInvalidChunkError: "Either the chunk does not exist in the file or it is not supported by the file.",
    kAudioFileDoesNotAllow64BitDataSizeError: "The file offset was too large for the file type. The AIFF and WAVE file format types have 32-bit file size limits.",
    kAudioFileInvalidPacketOffsetError: "A packet offset was past the end of the file, or not at the end of the file when a VBR format was written, or a corrupt packet size was read when the packet table was built.",
    kAudioFileInvalidFileError: "The file is malformed, or otherwise not a valid instance of an audio file of its type.",
    kAudioFileOperationNotSupportedError: "The operation cannot be performed.",
}


class CoreAudioException(Exception):
    """A stub to mark a problem as a CoreAudio issue.

    Ideally, there would be appropriate subclasses of typical
    Python exceptions for specific issues, For example:

    * kAudio_FileNotFoundError -> OSError (The typical Python file read error
    * kAudioFileInvalidChunkError -> ValueError (Invalid data)
    """
    ...


def err_check(err: int) -> None:
    """Raise an exception of somethings wrong, otherwise return None.

    Raises:
         CoreAudioException
    """
    if err != 0:
        raise CoreAudioException(err, err_str_db.get(err, "Unknown Error"))



import ctypes
from ctypes.wintypes import HANDLE, BYTE, HWND, BOOL, UINT, LONG, WORD, DWORD, WCHAR, LPVOID

lib = ctypes.windll.wintab32

FIX32 = DWORD
WTPKT = DWORD
HCTX = HANDLE  # CONTEXT HANDLE

LCNAMELEN = 40


class AXIS(ctypes.Structure):
    _fields_ = (
        ('axMin', LONG),
        ('axMax', LONG),
        ('axUnits', UINT),
        ('axResolution', FIX32)
    )

    def get_scale(self):
        return 1 / float(self.axMax - self.axMin)

    def get_bias(self):
        return -self.axMin


class ORIENTATION(ctypes.Structure):
    _fields_ = (
        ('orAzimuth', ctypes.c_int),
        ('orAltitude', ctypes.c_int),
        ('orTwist', ctypes.c_int)
    )


class ROTATION(ctypes.Structure):
    _fields_ = (
        ('roPitch', ctypes.c_int),
        ('roRoll', ctypes.c_int),
        ('roYaw', ctypes.c_int),
    )


class LOGCONTEXT(ctypes.Structure):
    _fields_ = (
        ('lcName', WCHAR * LCNAMELEN),
        ('lcOptions', UINT),
        ('lcStatus', UINT),
        ('lcLocks', UINT),
        ('lcMsgBase', UINT),
        ('lcDevice', UINT),
        ('lcPktRate', UINT),
        ('lcPktData', WTPKT),
        ('lcPktMode', WTPKT),
        ('lcMoveMask', WTPKT),
        ('lcBtnDnMask', DWORD),
        ('lcBtnUpMask', DWORD),
        ('lcInOrgX', LONG),
        ('lcInOrgY', LONG),
        ('lcInOrgZ', LONG),
        ('lcInExtX', LONG),
        ('lcInExtY', LONG),
        ('lcInExtZ', LONG),
        ('lcOutOrgX', LONG),
        ('lcOutOrgY', LONG),
        ('lcOutOrgZ', LONG),
        ('lcOutExtX', LONG),
        ('lcOutExtY', LONG),
        ('lcOutExtZ', LONG),
        ('lcSensX', FIX32),
        ('lcSensY', FIX32),
        ('lcSensZ', FIX32),
        ('lcSysMode', BOOL),
        ('lcSysOrgX', ctypes.c_int),
        ('lcSysOrgY', ctypes.c_int),
        ('lcSysExtX', ctypes.c_int),
        ('lcSysExtY', ctypes.c_int),
        ('lcSysSensX', FIX32),
        ('lcSysSensY', FIX32),
    )


class TILT(ctypes.Structure):  # 1.1
    _fields_ = (
        ('tiltX', ctypes.c_int),
        ('tiltY', ctypes.c_int),
    )


class EXTENSIONBASE(ctypes.Structure):  # 1.4
    _fields_ = (
        ('nContext', HCTX),  # Specifies the Wintab context to which these properties apply.
        ('nStatus', UINT),  # Status of setting/getting properties.
        ('nTime', DWORD),  # Timestamp applied to property transaction.
        ('nSerialNumber', UINT),  # Reserved - not use
    )


class EXPKEYSDATA(ctypes.Structure):  # 1.4
    _fields_ = (
        ('nTablet', BYTE),  # Tablet index where control is found.
        ('nControl', BYTE),  # Zero-based control index.
        ('nLocation', BYTE),  # Zero-based index indicating side of tablet where control found (0 = left, 1 = right).
        ('nReserved', BYTE),  # Reserved - not used
        ('nState', DWORD)  # Indicates Express Key button press (1 = pressed, 0 = released)
    )


class SLIDERDATA(ctypes.Structure):  # 1.4
    _fields_ = (
        ('nTablet', BYTE),  # Tablet index where control is found.
        ('nControl', BYTE),  # Zero-based control index.
        ('nMode', BYTE),  # Zero-based current active mode of control. Mode selected by control's toggle button.
        ('nReserved', BYTE),  # Reserved - not used
        ('nPosition', DWORD)  # An integer representing the position of the user's finger on the control.
        # When there is no finger on the control, this value is negative.
    )


class EXTPROPERTY(ctypes.Structure):  # 1.4
    _fields_ = (
        ('version', BYTE),  # Structure version, 0 for now
        ('tabletIndex', BYTE),  # 0-based index for tablet
        ('controlIndex', BYTE),  # 0-based index for control
        ('functionIndex', BYTE),  # 0-based index for control's sub-function
        ('propertyID', WORD),  # property ID
        ('reserved', WORD),  # DWORD-alignment filler
        ('dataSize', DWORD),  # number of bytes in data[] buffer
        ('data', BYTE * 1),  # raw data
    )


# Custom packet format with fields
#   PK_CHANGED
#   PK_CURSOR
#   PK_BUTTONS
#   PK_X
#   PK_Y
#   PK_Z
#   PK_NORMAL_PRESSURE
#   PK_TANGENT_PRESSURE
#   PK_ORIENTATION (check for tilt extension instead)?
class PACKET(ctypes.Structure):
    _fields_ = (
        ('pkChanged', WTPKT),
        ('pkCursor', UINT),
        ('pkButtons', DWORD),
        ('pkX', LONG),
        ('pkY', LONG),
        ('pkZ', LONG),
        ('pkNormalPressure', UINT),
        ('pkTangentPressure', UINT),
        ('pkOrientation', ORIENTATION),
    )


class PACKETEXT(ctypes.Structure):
    _fields_ = (
        ('pkBase', EXTENSIONBASE),  # Extension control properties common to all control types.
        ('pkExpKeys', EXPKEYSDATA),  # Extension data for one Express Key.
        ('pkTouchStrip', SLIDERDATA),  # Extension data for one Touch Strip.
        ('pkTouchRing', SLIDERDATA)  # Extension data for one Touch Ring.
    )


PK_CONTEXT = 0x0001  # reporting context
PK_STATUS = 0x0002  # status bits
PK_TIME = 0x0004  # time stamp
PK_CHANGED = 0x0008  # change bit vector
PK_SERIAL_NUMBER = 0x0010  # packet serial number
PK_CURSOR = 0x0020  # reporting cursor
PK_BUTTONS = 0x0040  # button information
PK_X = 0x0080  # x axis
PK_Y = 0x0100  # y axis
PK_Z = 0x0200  # z axis
PK_NORMAL_PRESSURE = 0x0400  # normal or tip pressure
PK_TANGENT_PRESSURE = 0x0800  # tangential or barrel pressure
PK_ORIENTATION = 0x1000  # orientation info: tilts
PK_ROTATION = 0x2000  # rotation info; 1.1

TU_NONE = 0
TU_INCHES = 1
TU_CENTIMETERS = 2
TU_CIRCLE = 3

# messages
WT_DEFBASE = 0x7ff0
WT_MAXOFFSET = 0xf
WT_PACKET = 0  # remember to add base
WT_CTXOPEN = 1
WT_CTXCLOSE = 2
WT_CTXUPDATE = 3
WT_CTXOVERLAP = 4
WT_PROXIMITY = 5
WT_INFOCHANGE = 6
WT_CSRCHANGE = 7
WT_PACKETEXT = 8

# system button assignment values 
SBN_NONE = 0x00
SBN_LCLICK = 0x01
SBN_LDBLCLICK = 0x02
SBN_LDRAG = 0x03
SBN_RCLICK = 0x04
SBN_RDBLCLICK = 0x05
SBN_RDRAG = 0x06
SBN_MCLICK = 0x07
SBN_MDBLCLICK = 0x08
SBN_MDRAG = 0x09

# for Pen Windows 
SBN_PTCLICK = 0x10
SBN_PTDBLCLICK = 0x20
SBN_PTDRAG = 0x30
SBN_PNCLICK = 0x40
SBN_PNDBLCLICK = 0x50
SBN_PNDRAG = 0x60
SBN_P1CLICK = 0x70
SBN_P1DBLCLICK = 0x80
SBN_P1DRAG = 0x90
SBN_P2CLICK = 0xA0
SBN_P2DBLCLICK = 0xB0
SBN_P2DRAG = 0xC0
SBN_P3CLICK = 0xD0
SBN_P3DBLCLICK = 0xE0
SBN_P3DRAG = 0xF0

HWC_INTEGRATED = 0x0001
HWC_TOUCH = 0x0002
HWC_HARDPROX = 0x0004
HWC_PHYSID_CURSORS = 0x0008  # 1.1

CRC_MULTIMODE = 0x0001  # 1.1
CRC_AGGREGATE = 0x0002  # 1.1
CRC_INVERT = 0x0004  # 1.1

WTI_INTERFACE = 1
IFC_WINTABID = 1
IFC_SPECVERSION = 2
IFC_IMPLVERSION = 3
IFC_NDEVICES = 4
IFC_NCURSORS = 5
IFC_NCONTEXTS = 6
IFC_CTXOPTIONS = 7
IFC_CTXSAVESIZE = 8
IFC_NEXTENSIONS = 9
IFC_NMANAGERS = 10
IFC_MAX = 10

WTI_STATUS = 2
STA_CONTEXTS = 1
STA_SYSCTXS = 2
STA_PKTRATE = 3
STA_PKTDATA = 4
STA_MANAGERS = 5
STA_SYSTEM = 6
STA_BUTTONUSE = 7
STA_SYSBTNUSE = 8
STA_MAX = 8

WTI_DEFCONTEXT = 3
WTI_DEFSYSCTX = 4
WTI_DDCTXS = 400  # 1.1
WTI_DSCTXS = 500  # 1.1
CTX_NAME = 1
CTX_OPTIONS = 2
CTX_STATUS = 3
CTX_LOCKS = 4
CTX_MSGBASE = 5
CTX_DEVICE = 6
CTX_PKTRATE = 7
CTX_PKTDATA = 8
CTX_PKTMODE = 9
CTX_MOVEMASK = 10
CTX_BTNDNMASK = 11
CTX_BTNUPMASK = 12
CTX_INORGX = 13
CTX_INORGY = 14
CTX_INORGZ = 15
CTX_INEXTX = 16
CTX_INEXTY = 17
CTX_INEXTZ = 18
CTX_OUTORGX = 19
CTX_OUTORGY = 20
CTX_OUTORGZ = 21
CTX_OUTEXTX = 22
CTX_OUTEXTY = 23
CTX_OUTEXTZ = 24
CTX_SENSX = 25
CTX_SENSY = 26
CTX_SENSZ = 27
CTX_SYSMODE = 28
CTX_SYSORGX = 29
CTX_SYSORGY = 30
CTX_SYSEXTX = 31
CTX_SYSEXTY = 32
CTX_SYSSENSX = 33
CTX_SYSSENSY = 34
CTX_MAX = 34

WTI_DEVICES = 100
DVC_NAME = 1
DVC_HARDWARE = 2
DVC_NCSRTYPES = 3
DVC_FIRSTCSR = 4
DVC_PKTRATE = 5
DVC_PKTDATA = 6
DVC_PKTMODE = 7
DVC_CSRDATA = 8
DVC_XMARGIN = 9
DVC_YMARGIN = 10
DVC_ZMARGIN = 11
DVC_X = 12
DVC_Y = 13
DVC_Z = 14
DVC_NPRESSURE = 15
DVC_TPRESSURE = 16
DVC_ORIENTATION = 17
DVC_ROTATION = 18  # 1.1
DVC_PNPID = 19  # 1.1
DVC_MAX = 19

WTI_CURSORS = 200
CSR_NAME = 1
CSR_ACTIVE = 2
CSR_PKTDATA = 3
CSR_BUTTONS = 4
CSR_BUTTONBITS = 5
CSR_BTNNAMES = 6
CSR_BUTTONMAP = 7
CSR_SYSBTNMAP = 8
CSR_NPBUTTON = 9
CSR_NPBTNMARKS = 10
CSR_NPRESPONSE = 11
CSR_TPBUTTON = 12
CSR_TPBTNMARKS = 13
CSR_TPRESPONSE = 14
CSR_PHYSID = 15  # 1.1
CSR_MODE = 16  # 1.1
CSR_MINPKTDATA = 17  # 1.1
CSR_MINBUTTONS = 18  # 1.1
CSR_CAPABILITIES = 19  # 1.1
CSR_TYPE = 20  # 1.2
CSR_MAX = 20

WTI_EXTENSIONS = 300
EXT_NAME = 1
EXT_TAG = 2
EXT_MASK = 3
EXT_SIZE = 4
EXT_AXES = 5
EXT_DEFAULT = 6
EXT_DEFCONTEXT = 7
EXT_DEFSYSCTX = 8
EXT_CURSORS = 9
EXT_MAX = 109  # Allow 100 cursors
CXO_SYSTEM = 0x0001
CXO_PEN = 0x0002
CXO_MESSAGES = 0x0004
CXO_MARGIN = 0x8000
CXO_MGNINSIDE = 0x4000
CXO_CSRMESSAGES = 0x0008  # 1.1

# context status values 
CXS_DISABLED = 0x0001
CXS_OBSCURED = 0x0002
CXS_ONTOP = 0x0004

# context lock values 
CXL_INSIZE = 0x0001
CXL_INASPECT = 0x0002
CXL_SENSITIVITY = 0x0004
CXL_MARGIN = 0x0008
CXL_SYSOUT = 0x0010
# packet status values 
TPS_PROXIMITY = 0x0001
TPS_QUEUE_ERR = 0x0002
TPS_MARGIN = 0x0004
TPS_GRAB = 0x0008
TPS_INVERT = 0x0010  # 1.1

TBN_NONE = 0
TBN_UP = 1
TBN_DOWN = 2
PKEXT_ABSOLUTE = 1
PKEXT_RELATIVE = 2

# Extension tags. 
WTX_OBT = 0  # Out of bounds tracking
WTX_FKEYS = 1  # Function keys
WTX_TILT = 2  # Raw Cartesian tilt; 1.1
WTX_CSRMASK = 3  # select input by cursor type; 1.1
WTX_XBTNMASK = 4  # Extended button mask; 1.1
WTX_EXPKEYS = 5  # ExpressKeys; 1.3  - DEPRECATED USE 2
WTX_TOUCHSTRIP = 6  # TouchStrips; 1.4
WTX_TOUCHRING = 7  # TouchRings; 1.4
WTX_EXPKEYS2 = 8  # ExpressKeys; 1.4

TABLET_PROPERTY_CONTROLCOUNT = 0  # UINT32: number of physical controls on tablet
TABLET_PROPERTY_FUNCCOUNT = 1  # UINT32: number of functions of control
TABLET_PROPERTY_AVAILABLE = 2  # BOOL: control/mode is available for override
TABLET_PROPERTY_MIN = 3  # UINT32: minimum value
TABLET_PROPERTY_MAX = 4  # UINT32: maximum value
TABLET_PROPERTY_OVERRIDE = 5  # BOOL: control is overridden
TABLET_PROPERTY_OVERRIDE_NAME = 6  # UTF-8: Displayable name when control is overridden
TABLET_PROPERTY_OVERRIDE_ICON = 7  # Image: Icon to show when control is overridden
TABLET_PROPERTY_ICON_WIDTH = 8  # UINT32: Pixel width of icon display
TABLET_PROPERTY_ICON_HEIGHT = 9  # UINT32: Pixel height of icon display
TABLET_PROPERTY_ICON_FORMAT = 10  # UINT32: UINT32: Pixel format of icon display (see TABLET_ICON_FMT_*)
TABLET_PROPERTY_LOCATION = 11  # UINT32: Physical location of control (see TABLET_LOC_*)

TABLET_LOC_LEFT = 0
TABLET_LOC_RIGHT = 1
TABLET_LOC_TOP = 2
TABLET_LOC_BOTTOM = 3
TABLET_LOC_TRANSDUCER = 4

lib.WTOpenW.restype = HCTX
lib.WTOpenW.argtypes = [HWND, ctypes.POINTER(LOGCONTEXT), BOOL]

lib.WTClose.restype = BOOL
lib.WTClose.argtypes = [HCTX]

lib.WTInfoW.restype = UINT
lib.WTInfoW.argtypes = [UINT, UINT, LPVOID]

lib.WTPacket.restype = BOOL
lib.WTPacket.argtypes = [HCTX, UINT, LPVOID]

lib.WTGetW.restype = BOOL
lib.WTGetW.argtypes = [HCTX, BOOL]

lib.WTExtGet.restype = BOOL
lib.WTExtGet.argtypes = [HCTX, UINT, LPVOID]

lib.WTExtSet.restype = BOOL
lib.WTExtSet.argtypes = [HCTX, UINT, LPVOID]



from __future__ import annotations

from typing import TYPE_CHECKING

import pyglet
from pyglet.gl import GLuint, glBindVertexArray, glDeleteVertexArrays, glGenVertexArrays

if TYPE_CHECKING:
    from ctypes import c_uint

    from pyglet.gl import Context

__all__ = ['VertexArray']


class VertexArray:
    """OpenGL Vertex Array Object."""
    _context: Context | None
    _id: c_uint

    def __init__(self) -> None:
        """Create an instance of a Vertex Array object."""
        self._context = pyglet.gl.current_context
        self._id = GLuint()
        glGenVertexArrays(1, self._id)

    @property
    def id(self) -> int:
        return self._id.value

    def bind(self) -> None:
        glBindVertexArray(self._id)

    @staticmethod
    def unbind() -> None:
        glBindVertexArray(0)

    def delete(self) -> None:
        glDeleteVertexArrays(1, self._id)
        self._id = None

    __enter__ = bind

    def __exit__(self, *_) -> None:  # noqa: ANN002
        glBindVertexArray(0)

    def __del__(self) -> None:
        if self._id is not None:
            try:
                self._context.delete_vao(self.id)
                self._id = None
            except (ImportError, AttributeError):
                pass  # Interpreter is shutting down

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self._id.value})"



#!/usr/bin/env python
"""Load and display a GIF animation.

Usage::

    animation.py [<filename>]

If the filename is omitted, a sample animation is loaded
"""
# The dinosaur.gif file packaged alongside this script is in the public
# domain, it was obtained from http://www.gifanimations.com/.

import sys

import pyglet

if len(sys.argv) > 1:
    # Load the animation from file path.
    animation = pyglet.image.load_animation(sys.argv[1])
    texture_bin = pyglet.image.atlas.TextureBin()
    animation.add_to_texture_bin(texture_bin)
else:
    # Load animation from resource (this script's directory).
    animation = pyglet.resource.animation('dinosaur.gif')

window = pyglet.window.Window(width=animation.get_max_width(), height=animation.get_max_height())
sprite = pyglet.sprite.Sprite(animation)


# Set window background color to white.
pyglet.gl.glClearColor(1, 1, 1, 1)


@window.event
def on_draw():
    window.clear()
    sprite.draw()


pyglet.app.run()



import os.path

from pyglet.image import *
from pyglet.image.codecs import *

from PIL import Image


class PILImageDecoder(ImageDecoder):
    def get_file_extensions(self):
        # Only most common ones shown here
        return ['.bmp', '.cur', '.gif', '.ico', '.jpg', '.jpeg', '.pcx', '.png',
                '.tga', '.tif', '.tiff', '.xbm', '.xpm']

    # def get_animation_file_extensions(self):
    #     return ['.gif', '.ani']

    def decode(self, filename, file):
        if not file:
            file = open(filename, 'rb')

        try:
            image = Image.open(file)
        except Exception as e:
            raise ImageDecodeException('PIL cannot read %r: %s' % (filename or file, e))

        try:
            image = image.transpose(Image.FLIP_TOP_BOTTOM)
        except Exception as e:
            raise ImageDecodeException('PIL failed to transpose %r: %s' % (filename or file, e))

        # Convert bitmap and palette images to component
        if image.mode in ('1', 'P'):
            image = image.convert()

        if image.mode not in ('L', 'LA', 'RGB', 'RGBA'):
            raise ImageDecodeException('Unsupported mode "%s"' % image.mode)
        width, height = image.size

        return ImageData(width, height, image.mode, image.tobytes())


class PILImageEncoder(ImageEncoder):
    def get_file_extensions(self):
        # Most common only
        return ['.bmp', '.eps', '.gif', '.jpg', '.jpeg',
                '.pcx', '.png', '.ppm', '.tiff', '.xbm']

    def encode(self, image, filename, file):
        # File format is guessed from filename extension, otherwise defaults to PNG.
        pil_format = (filename and os.path.splitext(filename)[1][1:]) or 'png'

        if pil_format.lower() == 'jpg':
            pil_format = 'JPEG'

        image = image.get_image_data()
        fmt = image.format
        if fmt != 'RGB':
            # Only save in RGB or RGBA formats.
            fmt = 'RGBA'
        pitch = -(image.width * len(fmt))

        # fromstring is deprecated, replaced by frombytes in Pillow (PIL fork)
        # (1.1.7) PIL still uses it
        try:
            image_from_fn = getattr(Image, "frombytes")
        except AttributeError:
            image_from_fn = getattr(Image, "fromstring")
        pil_image = image_from_fn(fmt, (image.width, image.height), image.get_data(fmt, pitch))

        try:
            pil_image.save(file, pil_format)
        except Exception as e:
            raise ImageEncodeException(e)


def get_decoders():
    return [PILImageDecoder()]


def get_encoders():
    return [PILImageEncoder()]



#!/usr/bin/env python3

import pyglet
from pyglet.experimental import multitexture_sprite

window = pyglet.window.Window()
window.set_size(400, 400)

pyglet.resource.path = ['../../examples/resources']
pyglet.resource.reindex()

# Batching allows rendering groups of objects all at once instead of drawing one by one.
batch = pyglet.graphics.Batch()

brick_grid = pyglet.image.ImageGrid(pyglet.resource.image("Brick1Gray.png", atlas=False), 1, 6)
brick_gray = pyglet.image.Animation.from_image_sequence(brick_grid, 1.0 / 10.0)
# Load 2 animation images for our sprites
brick_grid = pyglet.image.ImageGrid(pyglet.resource.image("Brick1Blue.png"), 1, 6)
brick_animation = pyglet.image.Animation.from_image_sequence(brick_grid, 1.0 / 10.0)
crack_grid = pyglet.image.ImageGrid(pyglet.resource.image("Brick1Crack3.png"), 1, 3)
crack_animation = pyglet.image.Animation.from_image_sequence(crack_grid, 1.0, False)
shader_images = {'brick': brick_animation, 'crack': crack_animation}

sprite = multitexture_sprite.MultiTextureSprite(shader_images,
                                                x=0,
                                                y=0,
                                                batch=batch)
sprite2 = multitexture_sprite.MultiTextureSprite({'brick': brick_animation, 'crack': crack_animation},
                                                 x=100,
                                                 y=100,
                                                 batch=batch)

# Custom shaders can also be used with the multi-texture sprite.  You must follow the naming
# convention for samplers and other uniforms.
custom_vertex_source = """#version 150 core
    in vec3 translate;
    in vec4 colors;

    // The MultiTextureSprite class expects the texture coordinates for each texture
    // to follow the naming convention {name}_coords
    in vec3 kitten_coords;
    in vec3 logo_coords;

    in vec2 scale;
    in vec3 position;
    in float rotation;

    out vec4 vertex_colors;

    // The MultiTextureSprite class expects the texture coordinates for each texture
    // going to the fragment shader to follow the naming convention {name}_coords_frag
    out vec3 kitten_coords_frag;
    out vec3 logo_coords_frag;

    uniform WindowBlock
    {
        mat4 projection;
        mat4 view;
    } window;

    mat4 m_scale = mat4(1.0);
    mat4 m_rotation = mat4(1.0);
    mat4 m_translate = mat4(1.0);

    void main()
    {
        m_scale[0][0] = scale.x;
        m_scale[1][1] = scale.y;
        m_translate[3][0] = translate.x;
        m_translate[3][1] = translate.y;
        m_translate[3][2] = translate.z;
        m_rotation[0][0] =  cos(-radians(rotation));
        m_rotation[0][1] =  sin(-radians(rotation));
        m_rotation[1][0] = -sin(-radians(rotation));
        m_rotation[1][1] =  cos(-radians(rotation));

        gl_Position = window.projection * window.view * m_translate * m_rotation * m_scale * vec4(position, 1.0);

        vertex_colors = colors;
        kitten_coords_frag = kitten_coords;
        logo_coords_frag = logo_coords;
    }
"""

custom_fragment_source = """#version 150 core
    in vec4 vertex_colors;

    // The MultiTextureSprite class expects the texture coordinates for each texture
    // to follow the naming convention {name}_coords_frag in the fragment shader
    in vec3 kitten_coords_frag;
    in vec3 logo_coords_frag;

    out vec4 final_colors;

    // The MultiTextureSprite class expects the sampler for each texture
    // to follow the naming convention {name}
    uniform sampler2D kitten;
    uniform sampler2D logo;

    void main()
    {
        final_colors = texture(kitten, kitten_coords_frag.xy) * texture(logo, logo_coords_frag.xy) * vertex_colors;
    }
"""

logo = pyglet.resource.image("pyglet.png")
kitten = pyglet.resource.image("kitten.jpg")
multi_vert_shader = pyglet.graphics.shader.Shader(custom_vertex_source, 'vertex')
multi_frag_shader = pyglet.graphics.shader.Shader(custom_fragment_source, 'fragment')
multitex_shader_program = pyglet.graphics.shader.ShaderProgram(multi_vert_shader, multi_frag_shader)
sprite3 = multitexture_sprite.MultiTextureSprite({'kitten': kitten, 'logo': logo},
                                                 x=150,
                                                 y=150,
                                                 batch=batch,
                                                 program=multitex_shader_program)
sprite3.scale = 0.2

f = 0


@window.event
def on_draw():
    window.clear()
    batch.draw()


def crack(dt):
    """Change frames for the second layer.
    """
    global f
    f = (f + 1) % 3
    sprite.set_frame_index('crack', f)
    pyglet.clock.schedule_once(crack, 1.0)


def pause(dt):
    print(f"Current paused value of sprite2 {sprite2.paused}")
    sprite2.paused = not sprite2.paused
    pyglet.clock.schedule_once(pause, 3.5)


def swap_animation(dt):
    sprite.set_layer('brick', brick_gray)


pyglet.clock.schedule_once(crack, 1.0)
pyglet.clock.schedule_once(pause, 3.5)
pyglet.clock.schedule_once(swap_animation, 5)

pyglet.app.run()



"""Wrapper for include/libswscale/swscale.h
"""
from ctypes import POINTER, Structure
from ctypes import c_int
from ctypes import c_uint8, c_double

import pyglet.lib
from pyglet.util import debug_print
from . import compat

_debug = debug_print('debug_media')


swscale = pyglet.lib.load_library(
    'swscale',
    win32=('swscale-8', 'swscale-7', 'swscale-6', 'swscale-5'),
    darwin=('swscale.8', 'swscale.7', 'swscale.6', 'swscale.5')
)

swscale.swscale_version.restype = c_int

compat.set_version('swscale', swscale.swscale_version() >> 16)


SWS_FAST_BILINEAR = 1


class SwsContext(Structure):
    pass


class SwsFilter(Structure):
    pass


swscale.sws_getCachedContext.restype = POINTER(SwsContext)
swscale.sws_getCachedContext.argtypes = [POINTER(SwsContext),
                                         c_int, c_int, c_int, c_int,
                                         c_int, c_int, c_int,
                                         POINTER(SwsFilter), POINTER(SwsFilter),
                                         POINTER(c_double)]
swscale.sws_freeContext.argtypes = [POINTER(SwsContext)]
swscale.sws_scale.restype = c_int
swscale.sws_scale.argtypes = [POINTER(SwsContext),
                              POINTER(POINTER(c_uint8)),
                              POINTER(c_int),
                              c_int, c_int,
                              POINTER(POINTER(c_uint8)),
                              POINTER(c_int)]

__all__ = [
    'swscale',
    'SWS_FAST_BILINEAR',
    'SwsContext',
    'SwsFilter'
]



import pyglet
import random


WIDTH = 1280
HEIGHT = 720
TARGET_FPS = 60

window = pyglet.window.Window(WIDTH, HEIGHT, caption="Benchmark", resizable=True, vsync=False)
label = pyglet.text.Label(" ", x=5, y=5, color=(223, 223, 223, 255), dpi=80)
batch = pyglet.graphics.Batch()


grey_img = pyglet.image.SolidColorImagePattern((50, 50, 50, 150)).create_image(5, 5)
blocks = []


@window.event
def on_draw():
    window.clear()
    batch.draw()
    label.draw()


@window.event
def on_resize(width, height):
    global WIDTH, HEIGHT
    WIDTH, HEIGHT = width, height


def update(dt):

    fps = 1 / dt
    count = int(fps - TARGET_FPS)

    if count > 0:
        for i in range(count//2):
            x = random.uniform(0, WIDTH/2)
            y = random.uniform(0, HEIGHT)
            blocks.append(pyglet.sprite.Sprite(img=grey_img, x=x, y=y, batch=batch, subpixel=True))

    else:
        if blocks:
            blocks.pop(0)

    for block in blocks:
        block.x += dt

    if blocks:
        label.text = f"FPS: {round(fps)}, sprites:{len(blocks)}"


pyglet.clock.schedule_interval(update, 1/120)
pyglet.app.run(1/120)



"""Provides keyboard and mouse editing procedures for text layout.

Example usage::

    from pyglet import window
    from pyglet.text import layout, caret

    my_window = window.Window(...)
    my_layout = layout.IncrementalTextLayout(...)
    my_caret = caret.Caret(my_layout)
    my_window.push_handlers(my_caret)
"""
from __future__ import annotations

import re
import time
from typing import TYPE_CHECKING, Any, Pattern

from pyglet import clock, event
from pyglet.window import key
from pyglet.event import EventDispatcher

if TYPE_CHECKING:
    from pyglet.graphics import Batch
    from pyglet.window import Window
    from pyglet.text.layout import IncrementalTextLayout

class Caret(EventDispatcher):
    """Visible text insertion marker for `pyglet.text.layout.IncrementalTextLayout`.

    The caret is drawn as a single vertical bar at the document `position`
    on a text layout object.  If ``mark`` is not None, it gives the unmoving
    end of the current text selection.  The visible text selection on the
    layout is updated along with ``mark`` and ``position``.

    By default, the layout's graphics batch is used, so the caret does not need
    to be drawn explicitly.  Even if a different graphics batch is supplied,
    the caret will be correctly positioned and clipped within the layout.

    Updates to the document (and so the layout) are automatically propagated
    to the caret.

    If the window argument is supplied, the caret object dispatches the on_clipboard_copy event when copying text and copies the text.
    Pasting also works, which will dispatch the on_clipboard_paste event, and pastes the text to the current position of the caret, overriding selection.

    The caret object can be pushed onto a window event handler stack with
    ``Window.push_handlers``.  The caret will respond correctly to keyboard,
    text, mouse and activation events, including double- and triple-clicks.
    If the text layout is being used alongside other graphical widgets, a
    GUI toolkit will be needed to delegate keyboard and mouse events to the
    appropriate widget.  Pyglet does not provide such a toolkit at this stage.
    """
    _next_word_re: Pattern[str] = re.compile(r"(?<=\W)\w")
    _previous_word_re: Pattern[str] = re.compile(r"(?<=\W)\w+\W*$")
    _next_para_re: Pattern[str] = re.compile(r"\n", flags=re.DOTALL)
    _previous_para_re: Pattern[str] = re.compile(r"\n", flags=re.DOTALL)

    _position: int = 0

    _active: bool = True
    _visible: bool = True
    _blink_visible: bool = True
    _click_count: int = 0
    _click_time: float = 0

    #: Blink period, in seconds.
    PERIOD: float = 0.5

    #: Pixels to scroll viewport per mouse scroll wheel movement.
    #: Defaults to 12pt at 96dpi.
    SCROLL_INCREMENT: int = 12 * 96 // 72

    _mark: int | None = None
    _next_attributes: dict[str, Any]

    def __init__(self, layout: IncrementalTextLayout, batch: Batch | None = None,
                 color: tuple[int, int, int, int] = (0, 0, 0, 255), window: Window | None = None) -> None:
        """Create a caret for a layout.

        By default the layout's batch is used, so the caret does not need to
        be drawn explicitly.

        :Parameters:
            `layout` : `~pyglet.text.layout.IncrementalTextLayout`
                Layout to control.
            `batch` : `~pyglet.graphics.Batch`
                Graphics batch to add vertices to.
            `color` : (int, int, int, int)
                An RGBA or RGB tuple with components in the range [0, 255].
                RGB colors will be treated as having an opacity of 255.
            `window` : `~pyglet.window.Window`
                For the clipboard feature to work, a window object is needed to be passed in to access and set clipboard content.
        """
        from pyglet import gl
        self._layout = layout

        self._custom_batch = batch is not None
        self._batch = batch or layout.batch
        self._group = layout.foreground_decoration_group

        # Handle both 3 and 4 byte colors
        r, g, b, *a = color

        # The alpha value when not in a hidden blink state
        self._visible_alpha = a[0] if a else 255

        colors = r, g, b, self._visible_alpha, r, g, b, self._visible_alpha

        self._list = self._group.program.vertex_list(2, gl.GL_LINES, self._batch, self._group,
                                                     colors=("Bn", colors),
                                                     visible=("f", (1, 1)))
        self._ideal_x = None
        self._ideal_line = None
        self._next_attributes = {}

        self.visible = True

        self._window = window
        layout.push_handlers(self)

    @property
    def layout(self) -> IncrementalTextLayout:
        return self._layout

    @layout.setter
    def layout(self, layout: IncrementalTextLayout) -> None:
        if self._layout == layout and self._group == layout.group:
            return

        from pyglet.gl import GL_LINES
        self._layout = layout
        batch = self._batch if self._custom_batch else layout.batch
        self._group = layout.foreground_decoration_group
        self._batch.migrate(self._list, GL_LINES, self._group, batch)

    def delete(self) -> None:
        """Remove the caret from its batch.

        Also disconnects the caret from further layout events.
        """
        clock.unschedule(self._blink)
        self._list.delete()
        self._layout.remove_handlers(self)

    def _blink(self, dt: float) -> None:  # noqa: ARG002
        if self.PERIOD:
            self._blink_visible = not self._blink_visible

        if self._visible and self._active and self._blink_visible:
            alpha = self._visible_alpha
        else:
            alpha = 0

        # Only set the alpha rather than entire colors
        self._list.colors[3] = alpha
        self._list.colors[7] = alpha

    def _nudge(self) -> None:
        self.visible = True

    @property
    def visible(self) -> bool:
        """Caret visibility.

        The caret may be hidden despite this property due to the periodic blinking
        or by `on_deactivate` if the event handler is attached to a window.
        """
        return self._visible

    @visible.setter
    def visible(self, visible: bool) -> None:
        self._visible = visible
        clock.unschedule(self._blink)
        if visible and self._active and self.PERIOD:
            clock.schedule_interval(self._blink, self.PERIOD)
            self._blink_visible = False  # flipped immediately by next blink
        self._blink(0)

    @property
    def color(self) -> tuple[int, int, int, int]:
        """An RGBA tuple of the current caret color.

        When blinking off, the alpha channel will be set to ``0``.  The
        default caret color when visible is ``(0, 0, 0, 255)`` (opaque black).

        You may set the color to an RGBA or RGB color tuple.

        .. warning:: This setter can fail for a short time after layout / window init!

                     Use ``__init__``'s ``color`` keyword argument instead if you
                     run into this problem.

        Each color channel must be between 0 and 255, inclusive. If the color
        set to an RGB color, the previous alpha channel value will be used.
        """
        return self._list.colors[:4]

    @color.setter
    def color(self, color: tuple[int, int, int, int] | tuple[int, int, int]) -> None:
        r, g, b, *_a = color

        # Preserve alpha when setting an RGB color
        a = _a[0] if _a else self._list.colors[3]

        self._list.colors[:] = r, g, b, a, r, g, b, a

    @property
    def position(self) -> int:
        """Position of caret within document."""
        return self._position

    @position.setter
    def position(self, position: int) -> None:
        self._position = position
        self._next_attributes.clear()
        self._update()

    @property
    def mark(self) -> int:
        """Position of immovable end of text selection within document.

        An interactive text selection is determined by its immovable end (the
        caret's position when a mouse drag begins) and the caret's position, which
        moves interactively by mouse and keyboard input.

        This property is ``None`` when there is no selection.
        """
        return self._mark

    @mark.setter
    def mark(self, mark: int) -> None:
        self._mark = mark
        self._update(line=self._ideal_line)
        if mark is None:
            self._layout.set_selection(0, 0)

    @property
    def line(self) -> int:
        """Index of line containing the caret's position.

        When set, ``position`` is modified to place the caret on requested line
        while maintaining the closest possible X offset.
        """
        if self._ideal_line is not None:
            return self._ideal_line

        return self._layout.get_line_from_position(self._position)

    @line.setter
    def line(self, line: int) -> None:
        if self._ideal_x is None:
            self._ideal_x, _ = self._layout.get_point_from_position(self._position)
        self._position = self._layout.get_position_on_line(line, self._ideal_x)
        self._update(line=line, update_ideal_x=False)

    def get_style(self, attribute: str) -> Any:
        """Get the document's named style at the caret's current position.

        If there is a text selection and the style varies over the selection,
        `pyglet.text.document.STYLE_INDETERMINATE` is returned.

        Args:
            attribute:
                Name of style attribute to retrieve.  See
                :py:mod:`~pyglet.text.document` for a list of recognised attribute
                names.
        """
        if self._mark is None or self._mark == self._position:
            try:
                return self._next_attributes[attribute]
            except KeyError:
                return self._layout.document.get_style(attribute, self._position)

        start = min(self._position, self._mark)
        end = max(self._position, self._mark)
        return self._layout.document.get_style_range(attribute, start, end)

    def set_style(self, attributes: dict[str, Any]) -> None:
        """Set the document style at the caret's current position.

        If there is a text selection the style is modified immediately.
        Otherwise, the next text that is entered before the position is
        modified will take on the given style.

        Args:
            attributes:
                Dict mapping attribute names to style values.  See
                :py:mod:`~pyglet.text.document` for a list of recognised attribute
                names.

        """
        if self._mark is None or self._mark == self._position:
            self._next_attributes.update(attributes)
            return

        start = min(self._position, self._mark)
        end = max(self._position, self._mark)
        self._layout.document.set_style(start, end, attributes)

    def _delete_selection(self) -> None:
        start = min(self._mark, self._position)
        end = max(self._mark, self._position)
        self._position = start
        self._mark = None
        self._layout.document.delete_text(start, end)
        self._layout.set_selection(0, 0)

    def move_to_point(self, x: int, y: int) -> None:
        """Move the caret close to the given window coordinate.

        The `mark` will be reset to ``None``.
        """
        line = self._layout.get_line_from_point(x, y)
        self._mark = None
        self._layout.set_selection(0, 0)
        self._position = self._layout.get_position_on_line(line, x)
        self._update(line=line)
        self._next_attributes.clear()

    def select_to_point(self, x: int, y: int) -> None:
        """Move the caret close to the given window coordinate while maintaining the `mark`."""
        line = self._layout.get_line_from_point(x, y)
        self._position = self._layout.get_position_on_line(line, x)
        self._update(line=line)
        self._next_attributes.clear()

    def select_all(self) -> None:
        """Select all text in the document."""
        self._mark = 0
        self._position = len(self._layout.document.text)
        self._update()
        self._next_attributes.clear()

    def select_word(self, x: int, y: int) -> None:
        """Select the word at the given window coordinate."""
        line = self._layout.get_line_from_point(x, y)
        p = self._layout.get_position_on_line(line, x)
        match1 = self._previous_word_re.search(self._layout.document.text, 0, p + 1)
        if not match1:
            mark1 = 0
        else:
            mark1 = match1.start()
        self.mark = mark1

        match2 = self._next_word_re.search(self._layout.document.text, p)
        if not match2:
            mark2 = len(self._layout.document.text)
        else:
            mark2 = match2.start()

        self._position = mark2
        self._update(line=line)
        self._next_attributes.clear()

    def select_paragraph(self, x: int, y: int) -> None:
        """Select the paragraph at the given window coordinate."""
        line = self._layout.get_line_from_point(x, y)
        p = self._layout.get_position_on_line(line, x)
        self.mark = self._layout.document.get_paragraph_start(p)
        self._position = self._layout.document.get_paragraph_end(p)
        self._update(line=line)
        self._next_attributes.clear()

    def _update(self, line: int | None = None, update_ideal_x: bool = True) -> None:
        if line is None:
            line = self._layout.get_line_from_position(self._position)
            self._ideal_line = None
        else:
            self._ideal_line = line
        x, y = self._layout.get_point_from_position(self._position, line)
        z = self._layout.z
        if update_ideal_x:
            self._ideal_x = x

        x += self._layout.left
        y += self._layout.bottom + self._layout._get_content_height()  # noqa: SLF001

        if self._mark is not None:
            self._layout.set_selection(min(self._position, self._mark), max(self._position, self._mark))

        self._layout.ensure_line_visible(line)
        self._layout.ensure_x_visible(x)

        font = self._layout.document.get_font(max(0, self._position - 1))
        self._list.position[:] = [x, y + font.descent, z, x, y + font.ascent, z]

    def on_translation_update(self) -> None:
        self._list.translation[:] = (-self._layout.view_x, -self._layout.view_y, 0) * 2

    def on_layout_update(self) -> None:
        """Handler for the `IncrementalTextLayout.on_layout_update` event."""
        if self.position > len(self._layout.document.text):
            self.position = len(self._layout.document.text)
        self._update()

    def on_text(self, text: str) -> bool:
        """Handler for the `pyglet.window.Window.on_text` event.

        Caret keyboard handlers assume the layout always has keyboard focus.
        GUI toolkits should filter keyboard and text events by widget focus
        before invoking this handler.
        """
        if self._mark is not None:
            self._delete_selection()

        text = text.replace("\r", "\n")
        pos = self._position
        self._position += len(text)
        self._layout.document.insert_text(pos, text, self._next_attributes)
        self._nudge()
        return event.EVENT_HANDLED

    def on_text_motion(self, motion: int, select: bool = False) -> bool:
        """Handler for the `pyglet.window.Window.on_text_motion` event.

        Caret keyboard handlers assume the layout always has keyboard focus.
        GUI toolkits should filter keyboard and text events by widget focus
        before invoking this handler.
        """
        if motion == key.MOTION_BACKSPACE:
            if self.mark is not None:
                self._delete_selection()
            elif self._position > 0:
                self._position -= 1
                self._layout.document.delete_text(self._position, self._position + 1)
                self._update()
        elif motion == key.MOTION_DELETE:
            if self.mark is not None:
                self._delete_selection()
            elif self._position < len(self._layout.document.text):
                self._layout.document.delete_text(self._position, self._position + 1)

        if motion == key.MOTION_LEFT:
            if self.mark is not None and not select:
                self.position = self._layout._selection_start  # noqa: SLF001
            else:
                self.position = max(0, self.position - 1)
        elif motion == key.MOTION_RIGHT:
            if self.mark is not None and not select:
                self.position = self._layout._selection_end  # noqa: SLF001
            else:
                self.position = min(len(self._layout.document.text), self.position + 1)
        elif motion == key.MOTION_UP:
            self.line = max(0, self.line - 1)
        elif motion == key.MOTION_DOWN:
            line = self.line
            if line < self._layout.get_line_count() - 1:
                self.line = line + 1
        elif motion == key.MOTION_BEGINNING_OF_LINE:
            self.position = self._layout.get_position_from_line(self.line)
        elif motion == key.MOTION_END_OF_LINE:
            line = self.line
            if line < self._layout.get_line_count() - 1:
                self._position = self._layout.get_position_from_line(line + 1) - 1
                self._update(line)
            else:
                self.position = len(self._layout.document.text)
        elif motion == key.MOTION_BEGINNING_OF_FILE:
            self.position = 0
        elif motion == key.MOTION_END_OF_FILE:
            self.position = len(self._layout.document.text)
        elif motion == key.MOTION_NEXT_WORD:
            pos = self._position + 1
            m = self._next_word_re.search(self._layout.document.text, pos)
            if not m:
                self.position = len(self._layout.document.text)
            else:
                self.position = m.start()
        elif motion == key.MOTION_PREVIOUS_WORD:
            pos = self._position
            m = self._previous_word_re.search(self._layout.document.text, 0, pos)
            if not m:
                self.position = 0
            else:
                self.position = m.start()
        elif motion == key.MOTION_COPY and self._window:
            pos = self._position
            mark = self._mark
            if pos > mark:
                text = self._layout.document.text[mark:pos]
            else:
                text = self._layout.document.text[pos:mark]

            self._window.set_clipboard_text(text)
            self.dispatch_event("on_clipboard_copy", text)
        elif motion == key.MOTION_PASTE and self._window:
            if self._mark is not None:
                self._delete_selection()

            text = self._window.get_clipboard_text().replace("\r", "\n")
            pos = self._position
            self._position += len(text)
            self._layout.document.insert_text(pos, text, self._next_attributes)
            self._nudge()
            self.dispatch_event("on_clipboard_paste", text)
        if self._mark is not None and not select:
            self._mark = None
            self._layout.set_selection(0, 0)

        self._next_attributes.clear()
        self._nudge()
        return event.EVENT_HANDLED

    def on_text_motion_select(self, motion: int) -> bool:
        """Handler for the `pyglet.window.Window.on_text_motion_select` event.

        Caret keyboard handlers assume the layout always has keyboard focus.
        GUI toolkits should filter keyboard and text events by widget focus
        before invoking this handler.
        """
        if self.mark is None:
            self.mark = self.position
        self.on_text_motion(motion, True)
        return event.EVENT_HANDLED

    def on_mouse_scroll(self, x: float, y: float, scroll_x: float, scroll_y: float) -> bool:  # noqa: ARG002
        """Handler for the `pyglet.window.Window.on_mouse_scroll` event.

        Mouse handlers do not check the bounds of the coordinates: GUI
        toolkits should filter events that do not intersect the layout
        before invoking this handler.

        The layout viewport is scrolled by `SCROLL_INCREMENT` pixels per
        "click".
        """
        self._layout.view_x -= scroll_x * self.SCROLL_INCREMENT
        self._layout.view_y += scroll_y * self.SCROLL_INCREMENT
        return event.EVENT_HANDLED

    def on_mouse_press(self, x: int, y: int, button: int, modifiers: int) -> bool:  # noqa: ARG002
        """Handler for the `pyglet.window.Window.on_mouse_press` event.

        Mouse handlers do not check the bounds of the coordinates: GUI
        toolkits should filter events that do not intersect the layout
        before invoking this handler.

        This handler keeps track of the number of mouse presses within
        a short span of time and uses this to reconstruct double- and
        triple-click events for selecting words and paragraphs.  This
        technique is not suitable when a GUI toolkit is in use, as the active
        widget must also be tracked.  Do not use this mouse handler if
        a GUI toolkit is being used.
        """
        t = time.time()
        if t - self._click_time < 0.25:
            self._click_count += 1
        else:
            self._click_count = 1
        self._click_time = time.time()

        if self._click_count == 1:
            self.move_to_point(x, y)
        elif self._click_count == 2:
            self.select_word(x, y)
        elif self._click_count == 3:
            self.select_paragraph(x, y)
            self._click_count = 0

        self._nudge()
        return event.EVENT_HANDLED

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> bool:  # noqa: ARG002
        """Handler for the `pyglet.window.Window.on_mouse_drag` event.

        Mouse handlers do not check the bounds of the coordinates: GUI
        toolkits should filter events that do not intersect the layout
        before invoking this handler.
        """
        if self.mark is None:
            self.mark = self.position
        self.select_to_point(x, y)
        self._nudge()
        return event.EVENT_HANDLED

    def on_activate(self) -> bool:
        """Handler for the `pyglet.window.Window.on_activate` event.

        The caret is hidden when the window is not active.
        """
        self._active = True
        self.visible = self._active
        return event.EVENT_HANDLED

    def on_deactivate(self) -> bool:
        """Handler for the `pyglet.window.Window.on_deactivate` event.

        The caret is hidden when the window is not active.
        """
        self._active = False
        self.visible = self._active
        return event.EVENT_HANDLED

    def on_clipboard_copy(self, text: str) -> bool:
        """Dispatched when text is copied.
        This default handler does nothing.
        """
        return event.EVENT_HANDLED
    def on_clipboard_paste(self, text: str) -> bool:
        """Dispatched when text is pasted.
        The default handler does nothing.
        """
        return event.EVENT_HANDLED

Caret.register_event_type("on_clipboard_copy")
Caret.register_event_type("on_clipboard_paste")



import pyglet


class Keyboard:
    def __init__(self):
        """A VERY basic semi-realtime synthesizer."""
        self.window = pyglet.window.Window(720, 480)
        instructions = "Press keys on your keyboard to play notes."
        self.instructions = pyglet.text.Label(text=instructions, font_size=20, x=10, y=10)
        self.current_note = pyglet.text.Label(text="", font_size=33, x=50, y=200)

        self.c4_notes = {"C": 261.63, "C#": 277.183,
                         "D": 293.66, "D#": 311.127,
                         "E": 329.63,
                         "F": 349.23, "F#": 369.994,
                         "G": 392.00, "G#": 415.305,
                         "A": 440.00, "A#": 466.164,
                         "B": 493.88, "R": 0}

        self.key_map = {pyglet.window.key.S: "C#",
                        pyglet.window.key.D: "D#",
                        pyglet.window.key.G: "F#",
                        pyglet.window.key.H: "G#",
                        pyglet.window.key.J: "A#",
                        pyglet.window.key.L: "C#",
                        pyglet.window.key.SEMICOLON: "D#",
                        pyglet.window.key.Z: "C",
                        pyglet.window.key.X: "D",
                        pyglet.window.key.C: "E",
                        pyglet.window.key.V: "F",
                        pyglet.window.key.B: "G",
                        pyglet.window.key.N: "A",
                        pyglet.window.key.M: "B",
                        pyglet.window.key.COMMA: "C",
                        pyglet.window.key.PERIOD: "D",
                        pyglet.window.key.BACKSLASH: "E"}

        self.note_cache = {}

        @self.window.event
        def on_key_press(key, mod):
            try:
                self.play_note(self.c4_notes[self.key_map[key]])
                self.current_note.text = f"Current note: {self.key_map[key]}"
            except KeyError:
                pass

        @self.window.event
        def on_draw():
            self.window.clear()
            self.instructions.draw()
            self.current_note.draw()

    def play_note(self, frequency, length=0.6):
        if frequency in self.note_cache:
            note_wave = self.note_cache[frequency]
            note_wave.play()
        else:
            adsr = pyglet.media.synthesis.ADSREnvelope(0.05, 0.2, 0.1)
            note_wave = pyglet.media.StaticSource(
                pyglet.media.synthesis.Sawtooth(duration=length, frequency=frequency, envelope=adsr))
            self.note_cache[frequency] = note_wave
            note_wave.play()


if __name__ == "__main__":
    keyboard = Keyboard()
    pyglet.app.run()



"""
A basic first-person camera example.

This is ideal for inspecting 3D models/scenes and can be adapted
to your needs for first person games.

* Supports mouse and keyboard input with WASD + QE for up/down.
* Supports controller input for movement and rotation including left and right triggers
  to move up and down,
"""

from __future__ import annotations

import weakref

from math import radians, degrees

import pyglet

from pyglet.gl import glEnable, GL_DEPTH_TEST, GL_CULL_FACE
from pyglet.math import Vec2, Vec3, Mat4, clamp
from pyglet.window import key as _key


class FPSCamera:
    """A 3D projection "first person" camera example.

    Windows in pyglet contain `view` and `projection` matrices,
    which are shared with all default shaders in a `WindowBlock` UBO.
    This Camera class handles events from the mouse & keyboard and
    Controller, and updates these Window matrices.

    Note:  mouse input will be captured once you click on the Window,
    which sets the mouse as exclusive. Pressing ESC once will set
    the mouse as non-exclusive.

    Args:
        window (pyglet.window.Window): The window to attach the camera to.

    Keyword Args:
        position (Vec3, optional):
            The position of the camera. Defaults to Vec3(0, 0, 0).
        target (Vec3, optional):
            The target of the camera. Defaults to Vec3(0, 0, -1).
        near (float, optional):
            The near plane of the camera. Defaults to 0.1.
        far (float, optional):
            The far plane of the camera. Defaults to 1000.
        field_of_view (float, optional):
            The field of view of the camera in degrees. Defaults to 45.0 degrees.
    """
    UP = Vec3(0.0, 1.0, 0.0)
    ZERO = Vec3(0.0, 0.0, 0.0)

    def __init__(
        self,
        window: pyglet.window.Window,
        position: Vec3 | None = None,
        target: Vec3 | None = None,
        near: float = 0.1,
        far: float = 1000,
        field_of_view: float = 60.0,
    ) -> None:
        """Initialize the camera."""
        self._window = weakref.proxy(window)
        self.position = position or Vec3(0.0, 0.0, 0.0)
        self._exclusive_mouse = False

        # Frustum
        self._near = near
        self._far = far
        self._field_of_view = field_of_view

        # Camera speed
        self.walk_speed = 10.0  # Speed of translation
        self.look_speed = 10.0  # Speed of rotation

        # Pitch, yaw in degrees. We update the camera based on these values.
        self._pitch = 0.0
        self._yaw = -90.0
        self._roll = 0.0
        self._elevation = 0.0

        # Input from keyboard and controller
        self.keyboard_move = Vec2()
        self.mouse_look = Vec2()

        self.controller_move = Vec2()
        self.controller_look = Vec2()
        self.dead_zone = 0.1

        # Keyboard input maps
        self.input_map = {  # Look up direction based on key symbol
            _key.W: "forward",
            _key.S: "backward",
            _key.A: "left",
            _key.D: "right",
            _key.E: "up",
            _key.Q: "down",
        }
        # Stores the direction string and the state of the direction
        self.inputs = {direction: False for direction in self.input_map.values()}

        # Make the camera point to the initial target.
        # We point it down the negative z-axis if no target is provided.
        if target is None:
            target = position + Vec3(0.0, 0.0, -1.0)

        self.teleport(position, target)

        self._window.push_handlers(self)
        self.on_refresh(0.0)

    @property
    def pitch(self) -> float:
        """Get or set the pitch of the camera in degrees.

        Pitch is the rotation around the local x-axis meaning
        that a positive pitch will look up and a negative pitch
        will look down.

        These values are clamped between -85 and 85 degrees to
        mimic the behavior of most first-person cameras.
        """
        return self._pitch
    
    @pitch.setter
    def pitch(self, value: float) -> None:
        self._pitch = clamp(value, -85.0, 85.0)

    @property
    def yaw(self) -> float:
        """The yaw of the camera in degrees.

        This is the local rotation around the global y-axis
        meaning that a positive yaw will look to the right
        and a negative yaw will look to the left.
        """
        return self._yaw

    @yaw.setter
    def yaw(self, value: float) -> None:
        self._yaw = value

    @property
    def field_of_view(self) -> float:
        """Get or set the field of view of the camera in degrees.

        Setting the value will update the projection matrix.
        """
        return self._field_of_view

    @field_of_view.setter
    def field_of_view(self, value: float) -> None:
        self._field_of_view = value
        self._update_projection()

    @property
    def near(self) -> float:
        """Get or set the near plane of the camera.

        Setting the value will update the projection matrix.
        """
        return self._near

    @near.setter
    def near(self, value: float) -> None:
        self._near = value
        self._update_projection()

    @property
    def far(self) -> float:
        """Get or set the far plane of the camera.

        Setting the value will update the projection matrix.
        """
        return self._far

    @far.setter
    def far(self, value: float) -> None:
        self._far = value
        self._update_projection()

    def on_resize(self, width: int, height: int) -> bool:
        """Update the viewport and projection matrix on window resize."""
        self._window.viewport = (0, 0, *self._window.get_framebuffer_size())
        self._update_projection()
        return pyglet.event.EVENT_HANDLED

    def on_refresh(self, delta_time: float) -> None:
        """Called before the window content is drawn.

        Runs every frame applying the camera movement.
        """
        walk_speed = self.walk_speed * delta_time
        look_speed = self.look_speed * delta_time

        # Rotation - mouse
        if self.mouse_look:
            self.yaw += self.mouse_look.x * look_speed
            self.pitch += self.mouse_look.y * look_speed
            # Reset the relative mouse movement when done.
            self.mouse_look = Vec2()

        # Rotation - controller
        if self.controller_look:
            self.yaw += self.controller_look.x * look_speed * 20
            self.pitch += self.controller_look.y * look_speed * 20

        # Calculate the local forward, right and up vectors from pitch and yaw
        forward = Vec3.from_pitch_yaw(radians(self.pitch), radians(self.yaw))
        right = forward.cross(self.UP).normalize()
        up = right.cross(forward).normalize()
        translation = Vec3()

        # Translation - keyboard
        if self.keyboard_move:
            translation += forward * self.keyboard_move.y + right * self.keyboard_move.x

        # Translation - controller
        if self.controller_move:
            translation += forward * self.controller_move.y + right * self.controller_move.x

        self.position += (translation + up * self._elevation) * walk_speed

        # Look forward from the new position
        self._window.view = Mat4.look_at(self.position, self.position + forward, self.UP)

    def on_deactivate(self) -> None:
        """Reset the movement states when the window loses focus."""
        self.controller_look = Vec2()
        self.controller_move = Vec2()

    def teleport(self, position: Vec3, target: Vec3 | None = None) -> None:
        """Teleport the camera to a new position.

        An optional new view target can be provided. If no target is
        provided, the camera will look in the same direction as before.
        """
        if target is not None:
            direction = (target - self.position).normalize()
            pitch, yaw = direction.get_pitch_yaw()
            self.yaw = degrees(yaw)
            self.pitch = degrees(pitch)

        self.position = position

    # --- Mouse input ---

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> None:
        """Read the mouse input and update the camera's yaw and pitch."""
        if not self._exclusive_mouse:
            return

        self.mouse_look = Vec2(dx, dy)

    def on_mouse_press(self, x: int, y: int, button, modifiers) -> None:
        """Capture the mouse input when the window is clicked."""
        if not self._exclusive_mouse:
            self._exclusive_mouse = True
            self._window.set_exclusive_mouse(True)

    # --- Keyboard input ---

    def on_key_press(self, symbol: int, mod: int) -> bool:
        """Handle keyboard input."""
        if direction := self.input_map.get(symbol):
            self.inputs[direction] = True
            forward, backward, left, right, up, down = self.inputs.values()
            self.keyboard_move = Vec2(-float(left) + float(right), float(forward) + -float(backward)).normalize()
            self._elevation = float(up) + -float(down)
            return pyglet.event.EVENT_HANDLED

        if symbol == pyglet.window.key.ESCAPE:
            if not self._exclusive_mouse:
                pyglet.app.exit()
            self._exclusive_mouse = False
            self._window.set_exclusive_mouse(False)
            return pyglet.event.EVENT_HANDLED

        return False

    def on_key_release(self, symbol: int, mod: int) -> bool:
        """Handle keyboard input release."""
        if direction := self.input_map.get(symbol):
            self.inputs[direction] = False
            forward, backward, left, right, up, down = self.inputs.values()
            self.keyboard_move = Vec2(-float(left) + float(right), float(forward) + -float(backward)).normalize()
            self._elevation = float(up) + -float(down)
            return pyglet.event.EVENT_HANDLED

        return False

    # --- Controller input ---

    def on_stick_motion(self, _controller, stick: str, vector: Vec2):
        """Handle controller input.

        The left stick controls the camera position, and the right stick
        controls the camera rotation.
        """
        # Translation
        if stick == "leftstick":
            if vector.length() < self.dead_zone:
                self.controller_move = Vec2()
            else:
                self.controller_move = vector

        # Camera rotation
        if stick == "rightstick":
            if vector.length() >= self.dead_zone:
                self.controller_look = vector
            else:
                self.controller_look = Vec2()

    def on_trigger_motion(self, controller, trigger: str, value: float):
        """Handle the controller trigger input."""
        if trigger == "lefttrigger":
            self._elevation = -value
        if trigger == "righttrigger":
            self._elevation = value

    # -- Private methods --

    def _update_projection(self):
        """Update the projection matrix"""
        self._window.projection = Mat4.perspective_projection(
            self._window.aspect_ratio,
            z_near=self.near,
            z_far=self.far,
            fov=self.field_of_view,
        )


if __name__ == "__main__":
    window = pyglet.window.Window(resizable=True)
    batch = pyglet.graphics.Batch()

    @window.event
    def on_draw():
        window.clear()
        batch.draw()

    glEnable(GL_DEPTH_TEST)
    glEnable(GL_CULL_FACE)

    # These .obj files only have a single model in the scene:
    model_logo = pyglet.resource.scene("logo3d.obj").create_models(batch=batch)[0]
    model_box = pyglet.resource.scene("box.obj").create_models(batch=batch)[0]

    # Camera controls the global projection & view matrixes:
    camera = FPSCamera(window, position=Vec3(0.0, 0.0, 5.0))

    # If a controller is connected, use it:
    if controllers := pyglet.input.get_controllers():
        controller = controllers[0]
        controller.open()
        controller.push_handlers(camera)

    model_logo.matrix = Mat4.from_translation(Vec3(1.75, 0, 0))
    model_box.matrix = Mat4.from_translation(Vec3(-1.75, 0, 0))

    pyglet.app.run()



"""
Usage

    compare.py --reldir=relpath other_session

Builds a reports comparing the active session with other_session.

Outputs to samples_dir/relpath/comparison_<session>_<other_session>.txt
"""
import os
import sys

import extractors
import fs
from pyglet.media import instrumentation as ins
import mpexceptions
import reports


def main(relative_out_dir, other_session):
    try:
        pathserv = fs.get_path_info_for_active_session()
    except mpexceptions.ExceptionUndefinedSamplesDir:
        print("The env var 'pyglet_mp_samples_dir' is not defined.")
        return 1
    except mpexceptions.ExceptionNoSessionIsActive:
        print("*** Error, no session active.")
        return 1

    try:
        pathserv_other = fs.get_path_info_for_session(other_session)
    except mpexceptions.ExceptionNoSessionWithThatName:
        print("No session by that name")
        return 1

    outdir = os.path.join(fs.get_current_samples_dir(), relative_out_dir)
    if not os.path.isdir(outdir):
        os.mkdir(outdir)

    compare_sessions_to_file(pathserv, pathserv_other, outdir)
    return 0


def compare_sessions_to_file(pathserv, pathserv_other, outdir):
    """compares two sessions, outputs text to file
    outdir/comparison_<session>_<other_session>.txt
    """
    outfile = os.path.join(outdir, "comparison_%s_%s.txt" %
                           (pathserv.session, pathserv_other.session))
    text = compare_sessions_to_text(pathserv, pathserv_other)
    fs.txt_save(text, outfile)


def compare_sessions_to_text(pathserv, pathserv_other):
    """compares two sessions, returns text"""
    count_bads = ins.CountBads()
    comparison = extractors.CollectSessionComparisonData(pathserv,
                                                         pathserv_other,
                                                         count_bads.count_bads)
    text = reports.report_session_comparison(comparison)
    return text


def sysargs_to_mainargs():
    """builds main args from sys.argv"""
    relative_out_dir = None
    if len(sys.argv) > 1 and sys.argv[1].startswith("--"):
        a = sys.argv.pop(1)
        if a.startswith("--help"):
            print(__doc__)
            sys.exit(1)
        elif a.startswith("--reldir="):
            relative_out_dir = a[len("--reldir="):]
        else:
            print("*** Error, Unknown option:", a)
            print(__doc__)
            sys.exit(1)

    other_session = sys.argv[1]
    return relative_out_dir, other_session


if __name__ == "__main__":
    relative_out_dir, other_session = sysargs_to_mainargs()
    main(relative_out_dir, other_session)



import sys
import warnings

from ctypes import CFUNCTYPE, byref, c_void_p, c_int, c_ubyte, c_bool, c_uint32, c_uint64

import pyglet

from pyglet.event import EventDispatcher
from pyglet.input.base import Device, AbsoluteAxis, RelativeAxis, Button
from pyglet.input.base import Joystick, Controller, AppleRemote, ControllerManager
from pyglet.input.controller import get_mapping, create_guid

from pyglet.libs.darwin.cocoapy import CFSTR, CFIndex, CFTypeID, known_cftypes
from pyglet.libs.darwin.cocoapy import kCFRunLoopDefaultMode, CFAllocatorRef, cf
from pyglet.libs.darwin.cocoapy import cfset_to_set, cftype_to_value, cfarray_to_list


__LP64__ = (sys.maxsize > 2 ** 32)

# Uses the HID API introduced in Mac OS X version 10.5
# http://developer.apple.com/library/mac/#technotes/tn2007/tn2187.html
iokit = pyglet.lib.load_library(framework='IOKit')

# IOKit constants from
# /System/Library/Frameworks/IOKit.framework/Headers/hid/IOHIDKeys.h
kIOHIDOptionsTypeNone	     = 0x00
kIOHIDOptionsTypeSeizeDevice = 0x01

kIOHIDElementTypeInput_Misc      = 1
kIOHIDElementTypeInput_Button    = 2
kIOHIDElementTypeInput_Axis      = 3
kIOHIDElementTypeInput_ScanCodes = 4
kIOHIDElementTypeOutput          = 129
kIOHIDElementTypeFeature         = 257
kIOHIDElementTypeCollection      = 513

# /System/Library/Frameworks/IOKit.framework/Headers/hid/IOHIDUsageTables.h
kHIDPage_GenericDesktop	       = 0x01
kHIDPage_Consumer              = 0x0C
kHIDUsage_GD_SystemSleep       = 0x82
kHIDUsage_GD_SystemWakeUp      = 0x83
kHIDUsage_GD_SystemAppMenu     = 0x86
kHIDUsage_GD_SystemMenu	       = 0x89
kHIDUsage_GD_SystemMenuRight   = 0x8A
kHIDUsage_GD_SystemMenuLeft    = 0x8B
kHIDUsage_GD_SystemMenuUp      = 0x8C
kHIDUsage_GD_SystemMenuDown    = 0x8D
kHIDUsage_Csmr_Menu            = 0x40
kHIDUsage_Csmr_FastForward     = 0xB3
kHIDUsage_Csmr_Rewind          = 0xB4
kHIDUsage_Csmr_Eject	       = 0xB8
kHIDUsage_Csmr_Mute	           = 0xE2
kHIDUsage_Csmr_VolumeIncrement = 0xE9
kHIDUsage_Csmr_VolumeDecrement = 0xEA

IOReturn = c_int  # IOReturn.h
IOOptionBits = c_uint32  # IOTypes.h

# IOHIDKeys.h
IOHIDElementType = c_int
IOHIDElementCollectionType = c_int
IOHIDElementCookie = c_uint32 if __LP64__ else c_void_p

iokit.IOHIDDeviceClose.restype = IOReturn
iokit.IOHIDDeviceClose.argtypes = [c_void_p, IOOptionBits]

iokit.IOHIDDeviceConformsTo.restype = c_ubyte
iokit.IOHIDDeviceConformsTo.argtypes = [c_void_p, c_uint32, c_uint32]

iokit.IOHIDDeviceCopyMatchingElements.restype = c_void_p
iokit.IOHIDDeviceCopyMatchingElements.argtypes = [c_void_p, c_void_p, IOOptionBits]

iokit.IOHIDDeviceGetProperty.restype = c_void_p
iokit.IOHIDDeviceGetProperty.argtypes = [c_void_p, c_void_p]

iokit.IOHIDDeviceGetTypeID.restype = CFTypeID
iokit.IOHIDDeviceGetTypeID.argtypes = []

iokit.IOHIDDeviceGetValue.restype = IOReturn
iokit.IOHIDDeviceGetValue.argtypes = [c_void_p, c_void_p, c_void_p]

iokit.IOHIDDeviceOpen.restype = IOReturn
iokit.IOHIDDeviceOpen.argtypes = [c_void_p, IOOptionBits]

iokit.IOHIDDeviceRegisterInputValueCallback.restype = None
iokit.IOHIDDeviceRegisterInputValueCallback.argtypes = [c_void_p, c_void_p, c_void_p]

iokit.IOHIDDeviceScheduleWithRunLoop.restype = None
iokit.IOHIDDeviceScheduleWithRunLoop.argtypes = [c_void_p, c_void_p, c_void_p]

iokit.IOHIDDeviceUnscheduleFromRunLoop.restype = None
iokit.IOHIDDeviceUnscheduleFromRunLoop.argtypes = [c_void_p, c_void_p, c_void_p]

iokit.IOHIDElementGetCollectionType.restype = IOHIDElementCollectionType
iokit.IOHIDElementGetCollectionType.argtypes = [c_void_p]

iokit.IOHIDElementGetCookie.restype = IOHIDElementCookie
iokit.IOHIDElementGetCookie.argtypes = [c_void_p]

iokit.IOHIDElementGetLogicalMax.restype = CFIndex
iokit.IOHIDElementGetLogicalMax.argtypes = [c_void_p]

iokit.IOHIDElementGetLogicalMin.restype = CFIndex
iokit.IOHIDElementGetLogicalMin.argtypes = [c_void_p]

iokit.IOHIDElementGetName.restype = c_void_p
iokit.IOHIDElementGetName.argtypes = [c_void_p]

iokit.IOHIDElementGetPhysicalMax.restype = CFIndex
iokit.IOHIDElementGetPhysicalMax.argtypes = [c_void_p]

iokit.IOHIDElementGetPhysicalMin.restype = CFIndex
iokit.IOHIDElementGetPhysicalMin.argtypes = [c_void_p]

iokit.IOHIDElementGetReportCount.restype = c_uint32
iokit.IOHIDElementGetReportCount.argtypes = [c_void_p]

iokit.IOHIDElementGetReportID.restype = c_uint32
iokit.IOHIDElementGetReportID.argtypes = [c_void_p]

iokit.IOHIDElementGetReportSize.restype = c_uint32
iokit.IOHIDElementGetReportSize.argtypes = [c_void_p]

iokit.IOHIDElementGetType.restype = IOHIDElementType
iokit.IOHIDElementGetType.argtypes = [c_void_p]

iokit.IOHIDElementGetTypeID.restype = CFTypeID
iokit.IOHIDElementGetTypeID.argtypes = []

iokit.IOHIDElementGetUnit.restype = c_uint32
iokit.IOHIDElementGetUnit.argtypes = [c_void_p]

iokit.IOHIDElementGetUnitExponent.restype = c_uint32
iokit.IOHIDElementGetUnitExponent.argtypes = [c_void_p]

iokit.IOHIDElementGetUsage.restype = c_uint32
iokit.IOHIDElementGetUsage.argtypes = [c_void_p]

iokit.IOHIDElementGetUsagePage.restype = c_uint32
iokit.IOHIDElementGetUsagePage.argtypes = [c_void_p]

iokit.IOHIDElementHasNullState.restype = c_bool
iokit.IOHIDElementHasNullState.argtypes = [c_void_p]

iokit.IOHIDElementHasPreferredState.restype = c_bool
iokit.IOHIDElementHasPreferredState.argtypes = [c_void_p]

iokit.IOHIDElementIsArray.restype = c_bool
iokit.IOHIDElementIsArray.argtypes = [c_void_p]

iokit.IOHIDElementIsNonLinear.restype = c_bool
iokit.IOHIDElementIsNonLinear.argtypes = [c_void_p]

iokit.IOHIDElementIsRelative.restype = c_bool
iokit.IOHIDElementIsRelative.argtypes = [c_void_p]

iokit.IOHIDElementIsVirtual.restype = c_bool
iokit.IOHIDElementIsVirtual.argtypes = [c_void_p]

iokit.IOHIDElementIsWrapping.restype = c_bool
iokit.IOHIDElementIsWrapping.argtypes = [c_void_p]

iokit.IOHIDManagerCreate.restype = c_void_p
iokit.IOHIDManagerCreate.argtypes = [CFAllocatorRef, IOOptionBits]

iokit.IOHIDManagerCopyDevices.restype = c_void_p
iokit.IOHIDManagerCopyDevices.argtypes = [c_void_p]

iokit.IOHIDManagerRegisterDeviceMatchingCallback.restype = None
iokit.IOHIDManagerRegisterDeviceMatchingCallback.argtypes = [c_void_p, c_void_p, c_void_p]

iokit.IOHIDManagerRegisterDeviceRemovalCallback.restype = None
iokit.IOHIDManagerRegisterDeviceRemovalCallback.argtypes = [c_void_p, c_void_p, c_void_p]

iokit.IOHIDManagerScheduleWithRunLoop.restype = c_void_p
iokit.IOHIDManagerScheduleWithRunLoop.argtypes = [c_void_p, c_void_p, c_void_p]

iokit.IOHIDManagerSetDeviceMatching.restype = None
iokit.IOHIDManagerSetDeviceMatching.argtypes = [c_void_p, c_void_p]

iokit.IOHIDValueGetElement.restype = c_void_p
iokit.IOHIDValueGetElement.argtypes = [c_void_p]

iokit.IOHIDValueGetIntegerValue.restype = CFIndex
iokit.IOHIDValueGetIntegerValue.argtypes = [c_void_p]

iokit.IOHIDValueGetLength.restype = CFIndex
iokit.IOHIDValueGetLength.argtypes = [c_void_p]

iokit.IOHIDValueGetTimeStamp.restype = c_uint64
iokit.IOHIDValueGetTimeStamp.argtypes = [c_void_p]

iokit.IOHIDValueGetTypeID.restype = CFTypeID
iokit.IOHIDValueGetTypeID.argtypes = []

# Callback function types
HIDManagerCallback = CFUNCTYPE(None, c_void_p, c_int, c_void_p, c_void_p)
HIDDeviceCallback = CFUNCTYPE(None, c_void_p, c_int, c_void_p)
HIDDeviceValueCallback = CFUNCTYPE(None, c_void_p, c_int, c_void_p, c_void_p)

######################################################################
# HID Class Wrappers

# Lookup tables cache python objects for the devices and elements so that
# we can avoid creating multiple wrapper objects for the same device.
_device_lookup = {}   # IOHIDDeviceRef to python HIDDevice object


class HIDValue:
    def __init__(self, value_ref):
        # Check that this is a valid IOHIDValue.
        assert value_ref
        assert cf.CFGetTypeID(value_ref) == iokit.IOHIDValueGetTypeID()
        self.value_ref = value_ref
        self.timestamp = iokit.IOHIDValueGetTimeStamp(value_ref)
        self.length = iokit.IOHIDValueGetLength(value_ref)
        if self.length <= 4:
            self.intvalue = iokit.IOHIDValueGetIntegerValue(value_ref)
        else:
            # Values may be byte data rather than integers.
            # e.g. the PS3 controller has a 39-byte HIDValue element.
            # We currently do not try to handle these cases.
            self.intvalue = None
        element_ref = c_void_p(iokit.IOHIDValueGetElement(value_ref))
        self.element = HIDDeviceElement.get_element(element_ref)


class HIDDevice:
    @classmethod
    def get_device(cls, device_ref):
        # device_ref is a c_void_p pointing to an IOHIDDeviceRef
        if device_ref.value in _device_lookup:
            return _device_lookup[device_ref.value]
        else:
            device = HIDDevice(device_ref)
            return device

    def __init__(self, device_ref):
        _device_lookup[device_ref.value] = self
        self.device_ref = device_ref
        # Set attributes from device properties.
        self.transport = self.get_property("Transport")
        self.vendorID = self.get_property("VendorID")
        self.vendorIDSource = self.get_property("VendorIDSource")
        self.productID = self.get_property("ProductID")
        self.versionNumber = self.get_property("VersionNumber")
        self.manufacturer = self.get_property("Manufacturer")
        self.product = self.get_property("Product")
        self.serialNumber = self.get_property("SerialNumber")  # always returns None; apple bug?
        self.locationID = self.get_property("LocationID")
        self.primaryUsage = self.get_property("PrimaryUsage")
        self.primaryUsagePage = self.get_property("PrimaryUsagePage")
        # Populate self.elements with our device elements.
        self.elements = self._get_elements()
        # Set up callback functions.
        self.value_observers = set()
        self.value_callback = self._register_input_value_callback()

    def get_guid(self):
        """Generate an SDL2 style GUID from the product guid."""
        # TODO: in what situation should 0x05 be used?
        # 0x03: USB
        # 0x05: Bluetooth
        bustype = 0x03
        vendor = self.vendorID or 0
        product = self.productID or 0
        version = self.versionNumber or 0
        name = self.product or ""
        return create_guid(bustype, vendor, product, version, name, 0, 0)

    def get_property(self, name):
        cfname = CFSTR(name)
        cfvalue = c_void_p(iokit.IOHIDDeviceGetProperty(self.device_ref, cfname))
        cf.CFRelease(cfname)
        return cftype_to_value(cfvalue)

    def open(self, exclusive_mode=False):
        if exclusive_mode:
            options = kIOHIDOptionsTypeSeizeDevice
        else:
            options = kIOHIDOptionsTypeNone
        return bool(iokit.IOHIDDeviceOpen(self.device_ref, options))

    def close(self):
        return bool(iokit.IOHIDDeviceClose(self.device_ref, kIOHIDOptionsTypeNone))

    def schedule_with_run_loop(self):
        iokit.IOHIDDeviceScheduleWithRunLoop(
            self.device_ref,
            c_void_p(cf.CFRunLoopGetCurrent()),
            kCFRunLoopDefaultMode)

    def unschedule_from_run_loop(self):
        iokit.IOHIDDeviceUnscheduleFromRunLoop(
            self.device_ref,
            c_void_p(cf.CFRunLoopGetCurrent()),
            kCFRunLoopDefaultMode)

    def _get_elements(self):
        cfarray = c_void_p(iokit.IOHIDDeviceCopyMatchingElements(self.device_ref, None, 0))
        if not cfarray:
            # requires "Security & Privacy / Input Monitoring", see #95
            return []
        elements = cfarray_to_list(cfarray)
        cf.CFRelease(cfarray)
        return elements

    # Page and usage IDs are from the HID usage tables located at
    # https://usb.org/sites/default/files/hut1_3_0.pdf
    def conforms_to(self, page, usage):
        return bool(iokit.IOHIDDeviceConformsTo(self.device_ref, page, usage))

    def is_pointer(self):
        return self.conforms_to(0x01, 0x01)

    def is_mouse(self):
        return self.conforms_to(0x01, 0x02)

    def is_joystick(self):
        return self.conforms_to(0x01, 0x04)

    def is_gamepad(self):
        return self.conforms_to(0x01, 0x05)

    def is_keyboard(self):
        return self.conforms_to(0x01, 0x06)

    def is_keypad(self):
        return self.conforms_to(0x01, 0x07)

    def is_multi_axis(self):
        return self.conforms_to(0x01, 0x08)

    def py_value_callback(self, context, result, sender, value):
        v = HIDValue(c_void_p(value))
        # Dispatch value changed message to all observers.
        for x in self.value_observers:
            if hasattr(x, 'device_value_changed'):
                x.device_value_changed(self, v)

    def _register_input_value_callback(self):
        value_callback = HIDDeviceValueCallback(self.py_value_callback)
        iokit.IOHIDDeviceRegisterInputValueCallback(self.device_ref, value_callback, None)
        return value_callback

    def add_value_observer(self, observer):
        self.value_observers.add(observer)

    def get_value(self, element):
        # If the device is not open, then returns None
        value_ref = c_void_p()
        iokit.IOHIDDeviceGetValue(self.device_ref, element.element_ref, byref(value_ref))
        if value_ref:
            return HIDValue(value_ref)
        else:
            return None

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.product})"


class HIDDeviceElement:

    @classmethod
    def get_element(cls, element_ref):
        # element_ref is a c_void_p pointing to an IOHIDDeviceElementRef
        return HIDDeviceElement(element_ref)

    def __init__(self, element_ref):
        self.element_ref = element_ref
        # Set element properties as attributes.
        self.cookie = iokit.IOHIDElementGetCookie(element_ref)
        self.type = iokit.IOHIDElementGetType(element_ref)
        if self.type == kIOHIDElementTypeCollection:
            self.collectionType = iokit.IOHIDElementGetCollectionType(element_ref)
        else:
            self.collectionType = None
        self.usagePage = iokit.IOHIDElementGetUsagePage(element_ref)
        self.usage = iokit.IOHIDElementGetUsage(element_ref)
        self.isVirtual = bool(iokit.IOHIDElementIsVirtual(element_ref))
        self.isRelative = bool(iokit.IOHIDElementIsRelative(element_ref))
        self.isWrapping = bool(iokit.IOHIDElementIsWrapping(element_ref))
        self.isArray = bool(iokit.IOHIDElementIsArray(element_ref))
        self.isNonLinear = bool(iokit.IOHIDElementIsNonLinear(element_ref))
        self.hasPreferredState = bool(iokit.IOHIDElementHasPreferredState(element_ref))
        self.hasNullState = bool(iokit.IOHIDElementHasNullState(element_ref))
        self.name = cftype_to_value(iokit.IOHIDElementGetName(element_ref))
        self.reportID = iokit.IOHIDElementGetReportID(element_ref)
        self.reportSize = iokit.IOHIDElementGetReportSize(element_ref)
        self.reportCount = iokit.IOHIDElementGetReportCount(element_ref)
        self.unit = iokit.IOHIDElementGetUnit(element_ref)
        self.unitExponent = iokit.IOHIDElementGetUnitExponent(element_ref)
        self.logicalMin = iokit.IOHIDElementGetLogicalMin(element_ref)
        self.logicalMax = iokit.IOHIDElementGetLogicalMax(element_ref)
        self.physicalMin = iokit.IOHIDElementGetPhysicalMin(element_ref)
        self.physicalMax = iokit.IOHIDElementGetPhysicalMax(element_ref)


class HIDManager(EventDispatcher):
    def __init__(self):
        """Create an instance of an HIDManager."""
        self.manager_ref = c_void_p(iokit.IOHIDManagerCreate(None, kIOHIDOptionsTypeNone))
        self.schedule_with_run_loop()

        self.devices = self._get_devices()
        self.matching_callback = self._register_matching_callback()
        self.removal_callback = self._register_removal_callback()

    def _get_devices(self):
        try:
            # Tell manager that we are willing to match *any* device.
            # (Alternatively, we could restrict by device usage, or usage page.)
            iokit.IOHIDManagerSetDeviceMatching(self.manager_ref, None)
            # Copy the device set and convert it to python.
            cfset = c_void_p(iokit.IOHIDManagerCopyDevices(self.manager_ref))
            devices = cfset_to_set(cfset)
            cf.CFRelease(cfset)
        except:
            return set()
        return devices

    def open(self):
        iokit.IOHIDManagerOpen(self.manager_ref, kIOHIDOptionsTypeNone)

    def close(self):
        iokit.IOHIDManagerClose(self.manager_ref, kIOHIDOptionsTypeNone)

    def schedule_with_run_loop(self):
        iokit.IOHIDManagerScheduleWithRunLoop(
            self.manager_ref,
            c_void_p(cf.CFRunLoopGetCurrent()),
            kCFRunLoopDefaultMode)

    def unschedule_from_run_loop(self):
        iokit.IOHIDManagerUnscheduleFromRunLoop(
            self.manager_ref,
            c_void_p(cf.CFRunLoopGetCurrent()),
            kCFRunLoopDefaultMode)

    # Device add/remove callbacks:

    def _py_matching_callback(self, context, result, sender, device):
        d = HIDDevice.get_device(c_void_p(device))
        if d not in self.devices:
            self.devices.add(d)
            self.dispatch_event("on_connect", d)

    def _register_matching_callback(self):
        matching_callback = HIDManagerCallback(self._py_matching_callback)
        iokit.IOHIDManagerRegisterDeviceMatchingCallback(self.manager_ref, matching_callback, None)
        return matching_callback

    def _py_removal_callback(self, context, result, sender, device):
        d = HIDDevice.get_device(c_void_p(device))
        d.close()
        if d in self.devices:
            self.devices.remove(d)
            self.dispatch_event("on_disconnect", d)

    def _register_removal_callback(self):
        removal_callback = HIDManagerCallback(self._py_removal_callback)
        iokit.IOHIDManagerRegisterDeviceRemovalCallback(self.manager_ref, removal_callback, None)
        return removal_callback


HIDManager.register_event_type('on_connect')
HIDManager.register_event_type('on_disconnect')


######################################################################
# Add conversion methods for IOHIDDevices and IOHIDDeviceElements
# to the list of known types used by cftype_to_value.
known_cftypes[iokit.IOHIDDeviceGetTypeID()] = HIDDevice.get_device
known_cftypes[iokit.IOHIDElementGetTypeID()] = HIDDeviceElement.get_element
######################################################################

# Pyglet interface to HID


_axis_names = {
    (0x01, 0x30): 'x',
    (0x01, 0x31): 'y',
    (0x01, 0x32): 'z',
    (0x01, 0x33): 'rx',
    (0x02, 0xc4): 'rx',
    (0x01, 0x34): 'ry',
    (0x02, 0xc5): 'ry',
    (0x01, 0x35): 'rz',
    (0x01, 0x38): 'wheel',
    (0x01, 0x39): 'hat',
}


_button_names = {
    (kHIDPage_GenericDesktop, kHIDUsage_GD_SystemSleep): 'sleep',
    (kHIDPage_GenericDesktop, kHIDUsage_GD_SystemWakeUp): 'wakeup',
    (kHIDPage_GenericDesktop, kHIDUsage_GD_SystemAppMenu): 'menu',
    (kHIDPage_GenericDesktop, kHIDUsage_GD_SystemMenu): 'select',
    (kHIDPage_GenericDesktop, kHIDUsage_GD_SystemMenuRight): 'right',
    (kHIDPage_GenericDesktop, kHIDUsage_GD_SystemMenuLeft): 'left',
    (kHIDPage_GenericDesktop, kHIDUsage_GD_SystemMenuUp): 'up',
    (kHIDPage_GenericDesktop, kHIDUsage_GD_SystemMenuDown): 'down',
    (kHIDPage_Consumer, kHIDUsage_Csmr_FastForward): 'right_hold',
    (kHIDPage_Consumer, kHIDUsage_Csmr_Rewind): 'left_hold',
    (kHIDPage_Consumer, kHIDUsage_Csmr_Menu): 'menu_hold',
    (0xff01, 0x23): 'select_hold',
    (kHIDPage_Consumer, kHIDUsage_Csmr_Eject): 'eject',
    (kHIDPage_Consumer, kHIDUsage_Csmr_Mute): 'mute',
    (kHIDPage_Consumer, kHIDUsage_Csmr_VolumeIncrement): 'volume_up',
    (kHIDPage_Consumer, kHIDUsage_Csmr_VolumeDecrement): 'volume_down'
}


class PygletDevice(Device):
    """
    Represents a device in the Pyglet input system.
    
    :param display: The display the device is attached to.
    :param device: The HIDDevice object.
    :param axis_filter: A list of usage_pages to include
                        (e.g. [0x1]).
    :param button_filter: A list of usage_pages to include
                          (e.g. [0x1]).
    """
    
    
    def __init__(self, display, device, axis_filter=None, button_filter=None):
        super().__init__(display=display, name=device.product)
        self.device = device
        self.device.add_value_observer(self)
        
        self._axis_filter = axis_filter
        self._button_filter = button_filter
        
        self._create_controls()

    def open(self, window=None, exclusive=False):
        super().open(window, exclusive)
        self.device.open(exclusive)
        self._set_initial_control_values()

    def close(self):
        super().close()
        self.device.close()

    def get_controls(self):
        return list(self._controls.values())

    def get_guid(self):
        return self.device.get_guid()

    def device_value_changed(self, hid_device, hid_value):
        # Called by device when input value changes.
        control = self._controls[hid_value.element.cookie]
        control.value = hid_value.intvalue

    def _create_controls(self):
        controls = []

        for element in self.device.elements:
            raw_name = element.name or '0x%x:%x' % (element.usagePage, element.usage)
            if element.type in (kIOHIDElementTypeInput_Misc, kIOHIDElementTypeInput_Axis):
                
                # apply axis filter to allow filtering for usage pages
                if self._axis_filter is not None and element.usagePage not in self._axis_filter:
                    continue
                
                name = _axis_names.get((element.usagePage, element.usage))
                if element.isRelative:
                    control = RelativeAxis(name, raw_name)
                else:
                    control = AbsoluteAxis(name, element.logicalMin, element.logicalMax, raw_name)
            elif element.type == kIOHIDElementTypeInput_Button:
                
                # apply button filter to allow filtering for usage pages
                if self._button_filter is not None and element.usagePage not in self._button_filter:
                    continue
                
                name = _button_names.get((element.usagePage, element.usage))
                control = Button(name, raw_name)
            else:
                continue

            control._cookie = element.cookie
            control._usage = element.usage
            controls.append(control)

        controls.sort(key=lambda c: c._usage)
        self._controls = {control._cookie: control for control in controls}

    def _set_initial_control_values(self):
        # Must be called AFTER the device has been opened.
        for element in self.device.elements:
            if element.cookie in self._controls:
                control = self._controls[element.cookie]
                hid_value = self.device.get_value(element)
                if hid_value:
                    control.value = hid_value.intvalue

    def __repr__(self):
        return f"{self.__class__.__name__}({self.device})"

######################################################################


_hid_manager = HIDManager()


class DarwinControllerManager(ControllerManager):

    def __init__(self, display=None):
        self._controllers = {}

        for device in _hid_manager.devices:
            if controller := _create_controller(device, display):
                self._controllers[device] = controller

        @_hid_manager.event
        def on_connect(hiddevice):
            if _controller := _create_controller(hiddevice, display):
                self._controllers[hiddevice] = _controller
                pyglet.app.platform_event_loop.post_event(self, 'on_connect', _controller)

        @_hid_manager.event
        def on_disconnect(hiddevice):
            if hiddevice in self._controllers:
                _controller = self._controllers[hiddevice]
                del self._controllers[hiddevice]
                pyglet.app.platform_event_loop.post_event(self, 'on_disconnect', _controller)

    def get_controllers(self):
        return list(self._controllers.values())


def get_devices(display=None):
    return [PygletDevice(display, device) for device in _hid_manager.devices]


def get_joysticks(display=None):
    return [Joystick(PygletDevice(display, device)) for device in _hid_manager.devices
            if device.is_joystick() or device.is_gamepad() or device.is_multi_axis()]


def get_apple_remote(display=None):
    for device in _hid_manager.devices:
        if device.product == 'Apple IR':
            return AppleRemote(PygletDevice(display, device))


def _create_controller(device, display):

    if not device.transport or not device.transport.upper() in ('USB', 'BLUETOOTH', 'BLUETOOTH LOW ENERGY'):
        return

    if device.is_joystick() or device.is_gamepad() or device.is_multi_axis():

        if mapping := get_mapping(device.get_guid()):
            return Controller(PygletDevice(display, device, axis_filter=[0x1, 0x2]), mapping)
        else:
            warnings.warn(f"Warning: {device} (GUID: {device.get_guid()}) "
                          f"has no controller mappings. Update the mappings in the Controller DB.")


def get_controllers(display=None):
    return [controller for controller in
            [_create_controller(device, display) for device in _hid_manager.devices
             if device.is_joystick() or device.is_gamepad() or device.is_multi_axis()]
            if controller is not None]



"""Factory functions to provide equivalent functionality as the Linux ioctl macros.

Ioctl macros are commonly used on Linux for interacting with character (char)
files. This module provides factory functions that create custom helper functions
to perform these same types of operations.

In pyglet, the input/evdev backend uses these functions for interacting with input
devices. For example, the helper functions are first defined::

    # Create a function for reading a HID descriptor size:
    HIDIOCGRDESCSIZE = _IOR(code='H', nr=0x01, struct=ctypes.c_uint)

    # Create a function for writing Force Feedback data:
    EVIOCSFF = _IOW(code='E', nr=0x80, struct=FFEventStruct)

    # Create a function to read a device name as a str:
    EVIOCGNAME = _IOR_str('E', 0x06)

These functions are then called to perform the operations::

    # Query the HID descriptor size (as uint):
    desc_size = HIDIOCGRDESCSIZE(fileno).value

    # Write a force feedback event (defined in a Structure):
    EVIOCSFF(fileno, ff_event_struct_instance)

    # Query the device name:
    name = EVIOCGNAME(fileno)


https://github.com/torvalds/linux/blob/master/include/uapi/asm-generic/ioctl.h

"""
from __future__ import annotations

from fcntl import ioctl as _ioctl
from ctypes import sizeof as _sizeof
from ctypes import create_string_buffer as _create_string_buffer
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ctypes import Structure, c_int, c_uint
    from typing import Callable, Union
    c_data = Union[type[Structure], c_int, c_uint]


_IOC_NRBITS = 8
_IOC_TYPEBITS = 8
_IOC_SIZEBITS = 14
_IOC_DIRBITS = 2

_IOC_NRSHIFT = 0
_IOC_TYPESHIFT = _IOC_NRSHIFT + _IOC_NRBITS
_IOC_SIZESHIFT = _IOC_TYPESHIFT + _IOC_TYPEBITS
_IOC_DIRSHIFT = _IOC_SIZESHIFT + _IOC_SIZEBITS

_IOC_NONE = 0
_IOC_WRITE = 1
_IOC_READ = 2


# To avoid name conflicts with Python, the following names are substituted:
# 'io_dir' instead of 'dir' to indicate the transfer direction (READ, WRITE, NONE).
# 'code' instead of 'type' to indicate the ioctl "magic number" ('H', 'E', etc.).


def _IOC(io_dir: _IOC_NONE | _IOC_READ | _IOC_WRITE, code: int, nr: int, size: int) -> int:
    return (io_dir << _IOC_DIRSHIFT) | (code << _IOC_TYPESHIFT) | (nr << _IOC_NRSHIFT) | (size << _IOC_SIZESHIFT)


def _IOR(code: str, nr: int, struct: c_data) -> Callable:
    request = _IOC(_IOC_READ, ord(code), nr, _sizeof(struct))

    def f(fileno, *args):
        buffer = struct(*args)
        _ioctl(fileno, request, buffer)
        return buffer

    return f


def _IOR_len(code: str, nr: int) -> Callable:

    def f(fileno, buffer):
        request = _IOC(_IOC_READ, ord(code), nr, _sizeof(buffer))
        _ioctl(fileno, request, buffer)
        return buffer

    return f


def _IOR_str(code: str, nr: int) -> Callable:
    g = _IOR_len(code, nr)

    def f(fileno, length=256):
        return g(fileno, _create_string_buffer(length)).value

    return f


def _IOW(code: str, nr: int, struct: c_data) -> Callable:
    request = _IOC(_IOC_WRITE, ord(code), nr, _sizeof(struct))

    def f(fileno, buffer):
        if not isinstance(buffer, struct):
            buffer = struct(buffer)
        _ioctl(fileno, request, buffer)

    return f


def _IOWR(code: str, nr: int, struct: c_data) -> Callable:
    request = _IOC(_IOC_READ | _IOC_WRITE, ord(code), nr, _sizeof(struct))

    def f(fileno, buffer):
        _ioctl(fileno, request, buffer)
        return buffer

    return f


def _IOWR_len(code: str, nr: int) -> Callable:

    def f(fileno, buffer):
        request = _IOC(_IOC_READ | _IOC_WRITE, ord(code), nr, _sizeof(buffer))
        _ioctl(fileno, request, buffer)
        return buffer

    return f



"""
Layout:

.                               (file home)
    f1.txt                          F1
    dir1/
        f2.txt                      F2
        dir1/
            f3.txt                  F3
        res.zip/
            f7.txt                  F7
            dir1/
                f8.txt              F8
                dir1/
                    f9.txt          F9
    dir2/
        f6.txt                      F6

"""

import os

import pytest

from pyglet import resource
from pyglet.util import asbytes


@pytest.fixture
def loader():
    script_home = os.path.dirname(__file__)
    return resource.Loader(script_home=script_home)


def test_base_path_only(loader):
    assert loader.file('f1.txt').read().strip() == asbytes('F1')


def test_blank_base_path(loader):
    loader.path = ['']
    assert loader.file('f1.txt').read().strip() == asbytes('F1')


def test_unused_paths(loader):
    loader.path = ['foo', 'bar', '.']
    assert loader.file('f1.txt').read().strip() == asbytes('F1')


def test_subfolder(loader):
    loader.path = ['dir1', 'dir2']
    assert loader.file('f2.txt').read().strip() == asbytes('F2')
    assert loader.file('f6.txt').read().strip() == asbytes('F6')


def test_subfolder_trailing_slash(loader):
    loader.path = ['dir1/', 'dir2/']
    assert loader.file('f2.txt').read().strip() == asbytes('F2')
    assert loader.file('f6.txt').read().strip() == asbytes('F6')


def test_sub_subfolder(loader):
    loader.path = ['dir1/dir1']
    assert loader.file('f3.txt').read().strip() == asbytes('F3')


def test_sub_subfolder_trailing_slash(loader):
    loader.path = ['dir1/dir1/']
    assert loader.file('f3.txt').read().strip() == asbytes('F3')


def test_zipfile(loader):
    loader.path = ['dir1/res.zip']
    assert loader.file('f7.txt').read().strip() == asbytes('F7')


def test_zipfile_trailing_slash(loader):
    loader.path = ['dir1/res.zip/']
    assert loader.file('f7.txt').read().strip() == asbytes('F7')


def test_zipfile_subdirs(loader):
    loader.path = ['dir1/res.zip/dir1', 'dir1/res.zip/dir1/dir1/']
    assert loader.file('f8.txt').read().strip() == asbytes('F8')
    assert loader.file('f9.txt').read().strip() == asbytes('F9')


# Expected Failures:

def test_no_path_exception(loader):
    loader.path = []
    pytest.raises(resource.ResourceNotFoundException, loader.file, 'f1.txt')


def test_resource_not_found(loader):
    pytest.raises(resource.ResourceNotFoundException, loader.file, 'foo')


def test_invalid_filename_format(loader):
    pytest.raises((AttributeError, TypeError), loader.file, ['foo'])



"""Testing the events"""
import gc
import types

from tests import mock

import pytest
import pyglet

from pyglet.event import EVENT_HANDLED, EVENT_UNHANDLED


@pytest.fixture
def dispatcher():
    """Event dispatcher to test"""
    dispatcher = pyglet.event.EventDispatcher()
    yield dispatcher
    # Let's remove previous handlers
    dispatcher._event_stack = ()
    try:
        del pyglet.event.EventDispatcher.event_types
    except AttributeError:
        pass


@pytest.fixture
def mock_handler(dispatcher):
    """Event handler mock."""
    # Make a mock behave like a function by replicating the FunctionType specs.
    # Function objects provide these attributes:
    # __doc__         documentation string
    # __name__        name with which this function was defined
    # __code__        code object containing compiled function bytecode
    # __defaults__    tuple of any default values for arguments
    # __globals__     global namespace in which this function was defined
    # __annotations__ dict of parameter annotations
    # __kwdefaults__  dict of keyword only parameters with defaults
    dispatcher.register_event_type('mock_event')
    mock_handler = mock.Mock(spec=types.FunctionType)
    mock_handler.__name__ = 'mock_event'
    return mock_handler


class ClassInstanceHandler:
    called = False
    args = None
    kwargs = None

    def mock_event(self, *args, **kwargs):
        self.called = True
        self.args = args
        self.kwargs = kwargs
        return EVENT_HANDLED

    def mock_event2(self, *args, **kwargs):
        self.called = True
        self.args = args
        self.kwargs = kwargs
        return EVENT_HANDLED



@pytest.fixture
def class_instance_handler(dispatcher):
    """Class event handler.

    If a class instance is given to push_handlers, and it implements methods with
    the same names as the events, those will become event handlers.
    """
    dispatcher.register_event_type('mock_event')
    return ClassInstanceHandler()


def test_register_event_type(dispatcher):
    dispatcher.register_event_type('mock_event')
    assert len(dispatcher.event_types) == 1


def test_push_handlers_args(dispatcher, mock_handler):
    dispatcher.push_handlers(mock_handler)
    result = dispatcher.dispatch_event('mock_event')
    assert result == EVENT_HANDLED
    assert mock_handler.called


def test_push_handlers_kwargs(dispatcher, mock_handler):
    dispatcher.push_handlers(mock_event=mock_handler)
    result = dispatcher.dispatch_event('mock_event')
    assert result == EVENT_HANDLED
    assert mock_handler.called


def test_push_handlers_instance(dispatcher, class_instance_handler):
    dispatcher.register_event_type('mock_event')
    dispatcher.push_handlers(class_instance_handler)
    result = dispatcher.dispatch_event('mock_event', 1, 2)
    assert result == EVENT_HANDLED
    assert class_instance_handler.called
    assert class_instance_handler.args == (1, 2)
    assert class_instance_handler.kwargs == {}


def test_push_handlers_not_setup(dispatcher):
    dispatcher.push_handlers()
    assert dispatcher._event_stack == [{}]


def test_set_handlers_args(dispatcher, mock_handler):
    dispatcher.set_handlers(mock_handler)
    result = dispatcher.dispatch_event('mock_event')
    assert result == EVENT_HANDLED
    assert mock_handler.called


def test_set_handlers_kwargs(dispatcher, mock_handler):
    dispatcher.set_handlers(mock_event=mock_handler)
    result = dispatcher.dispatch_event('mock_event')
    assert result == EVENT_HANDLED
    assert mock_handler.called


def test_set_handlers_not_setup(dispatcher):
    dispatcher.set_handlers()
    assert dispatcher._event_stack == [{}]


def test_set_handler_dispatch(dispatcher, mock_handler):
    dispatcher.set_handler('mock_event', mock_handler)
    result = dispatcher.dispatch_event('mock_event')
    assert result == EVENT_HANDLED
    assert mock_handler.called


def test_set_handler_not_setup(dispatcher, mock_handler):
    dispatcher.set_handler('mock_event', None)
    result = dispatcher.dispatch_event('mock_event')
    assert result is False
    assert not mock_handler.called


def test_pop_handlers(dispatcher, mock_handler):
    dispatcher.set_handler('mock_event', None)
    dispatcher.pop_handlers()
    with pytest.raises(AssertionError):
        dispatcher.pop_handlers()


def test_pop_handlers_not_setup(dispatcher):
    with pytest.raises(AssertionError):
        dispatcher.pop_handlers()


def test_remove_handlers_args(dispatcher, mock_handler):
    dispatcher.set_handler('mock_event', mock_handler)
    dispatcher.remove_handlers(mock_handler)
    result = dispatcher.dispatch_event('mock_event')
    assert result is False
    assert not mock_handler.called


def test_remove_handlers_kwargs(dispatcher, mock_handler):
    dispatcher.set_handler('mock_event', mock_handler)
    dispatcher.remove_handlers(mock_event=mock_handler)
    result = dispatcher.dispatch_event('mock_event')
    assert result is False
    assert not mock_handler.called


def test_remove_handlers_not_setup(dispatcher):
    assert dispatcher.remove_handlers() is None


def test_remove_handler(dispatcher, mock_handler):
    dispatcher.set_handler('mock_event', mock_handler)
    dispatcher.remove_handler('mock_event', mock_handler)
    result = dispatcher.dispatch_event('mock_event')
    assert result is False
    assert not mock_handler.called


def test_dispatch_unhandled_event(dispatcher):
    dispatcher.register_event_type('mock_event')
    with pytest.raises(AssertionError):
        dispatcher.dispatch_event('not_handled')


def test_dispatch_event_not_setup(dispatcher):
    with pytest.raises(AssertionError):
        dispatcher.dispatch_event('mock_event')


class DummyHandler:
    def mock_event(self):
        return True


def test_weakref_to_instance_method(dispatcher):
    import weakref
    dispatcher.register_event_type('mock_event')
    handler = DummyHandler()
    watcher = mock.Mock()
    weakref.finalize(handler, watcher)
    dispatcher.push_handlers(handler.mock_event)
    del handler
    gc.collect()    # ensure references are cleared
    assert watcher.called


def test_weakref_to_instance(dispatcher):
    import weakref
    dispatcher.register_event_type('mock_event')
    handler = DummyHandler()
    watcher = mock.Mock()
    weakref.finalize(handler, watcher)
    dispatcher.push_handlers(handler)
    del handler
    gc.collect()    # ensure references are cleared
    assert watcher.called


def test_weakref_deleted_when_instance_is_deleted(dispatcher):
    dispatcher.register_event_type('mock_event')
    handler = DummyHandler()
    dispatcher.push_handlers(handler.mock_event)
    del handler
    gc.collect()    # ensure references are cleared
    result = dispatcher.dispatch_event('mock_event')
    assert result is False



from __future__ import annotations

import warnings
from _ctypes import _Pointer
from ctypes import POINTER, byref, c_int, c_uint, cast
from typing import TYPE_CHECKING, NoReturn

from pyglet import gl
from pyglet.gl import glx
from pyglet.gl import glxext_arb
from pyglet.gl import glx_info
from pyglet.gl import glxext_mesa
from pyglet.gl import lib
from pyglet.gl.base import Config, DisplayConfig, Context

from pyglet.display.xlib import XlibCanvas
from pyglet.libs.x11 import xlib
from pyglet.libs.x11.xrender import XRenderFindVisualFormat

if TYPE_CHECKING:
    from pyglet.libs.x11.xlib import Display
    from .glx_info import GLXInfo


class XlibConfig(Config):  # noqa: D101

    def match(self, canvas: XlibCanvas) -> list[XlibDisplayConfig]:
        if not isinstance(canvas, XlibCanvas):
            msg = f'Canvas must be an instance of XlibCanvas {type(canvas)}'
            raise RuntimeError(msg)

        x_display = canvas.display._display  # noqa: SLF001
        x_screen = canvas.display.x_screen

        info = glx_info.GLXInfo(x_display)

        # Construct array of attributes
        attrs = []
        for name, value in self.get_gl_attributes():
            attr = XlibDisplayConfig.attribute_ids.get(name, None)
            if attr and value is not None:
                attrs.extend([attr, int(value)])

        attrs.extend([glx.GLX_X_RENDERABLE, True])
        attrs.extend([0, 0])  # attrib_list must be null terminated

        attrib_list = (c_int * len(attrs))(*attrs)

        elements = c_int()
        configs = glx.glXChooseFBConfig(x_display, x_screen, attrib_list, byref(elements))
        if not configs:
            return []

        configs = cast(configs, POINTER(glx.GLXFBConfig * elements.value)).contents

        result = [XlibDisplayConfig(canvas, info, c, self) for c in configs]

        # If we intend to have a transparent framebuffer.
        if self.transparent_framebuffer:
            result = [fb_cf for fb_cf in result if fb_cf.transparent]

        # Can't free array until all XlibGLConfig's are GC'd.  Too much
        # hassle, live with leak. XXX
        # xlib.XFree(configs)

        return result


class XlibDisplayConfig(DisplayConfig):  # noqa: D101
    glx_info: GLXInfo

    attribute_ids = {  # noqa: RUF012
        'buffer_size': glx.GLX_BUFFER_SIZE,
        'level': glx.GLX_LEVEL,  # Not supported
        'double_buffer': glx.GLX_DOUBLEBUFFER,
        'stereo': glx.GLX_STEREO,
        'aux_buffers': glx.GLX_AUX_BUFFERS,
        'red_size': glx.GLX_RED_SIZE,
        'green_size': glx.GLX_GREEN_SIZE,
        'blue_size': glx.GLX_BLUE_SIZE,
        'alpha_size': glx.GLX_ALPHA_SIZE,
        'depth_size': glx.GLX_DEPTH_SIZE,
        'stencil_size': glx.GLX_STENCIL_SIZE,
        'accum_red_size': glx.GLX_ACCUM_RED_SIZE,
        'accum_green_size': glx.GLX_ACCUM_GREEN_SIZE,
        'accum_blue_size': glx.GLX_ACCUM_BLUE_SIZE,
        'accum_alpha_size': glx.GLX_ACCUM_ALPHA_SIZE,

        'sample_buffers': glx.GLX_SAMPLE_BUFFERS,
        'samples': glx.GLX_SAMPLES,

        # Not supported in current pyglet API:
        'render_type': glx.GLX_RENDER_TYPE,
        'config_caveat': glx.GLX_CONFIG_CAVEAT,
        'transparent_type': glx.GLX_TRANSPARENT_TYPE,
        'transparent_index_value': glx.GLX_TRANSPARENT_INDEX_VALUE,
        'transparent_red_value': glx.GLX_TRANSPARENT_RED_VALUE,
        'transparent_green_value': glx.GLX_TRANSPARENT_GREEN_VALUE,
        'transparent_blue_value': glx.GLX_TRANSPARENT_BLUE_VALUE,
        'transparent_alpha_value': glx.GLX_TRANSPARENT_ALPHA_VALUE,

        # Used internally
        'x_renderable': glx.GLX_X_RENDERABLE,
    }

    def __init__(self, canvas: XlibCanvas, info: GLXInfo, fbconfig: glx.GLXFBConfig,  # noqa: D107
                 config: XlibConfig) -> None:
        super().__init__(canvas, config)

        self.glx_info = info
        self.fbconfig = fbconfig
        self.transparent = False

        for name, attr in self.attribute_ids.items():
            value = c_int()
            result = glx.glXGetFBConfigAttrib(canvas.display._display, self.fbconfig, attr,  # noqa: SLF001
                                              byref(value))
            if result >= 0:
                setattr(self, name, value.value)

        # If user intends for a transparent framebuffer, the visual info needs to be
        # queried for it. Even if a config supports alpha_size 8 and depth_size 32, there is no
        # guarantee the visual info supports that same configuration.
        if config.transparent_framebuffer:
            xvi_ptr = glx.glXGetVisualFromFBConfig(canvas.display._display, self.fbconfig)  # noqa: SLF001
            if xvi_ptr:
                self.transparent = self._is_visual_transparent(xvi_ptr.contents.visual)  # noqa: SLF001
                xlib.XFree(xvi_ptr)

    def _is_visual_transparent(self, visual: _Pointer[xlib.Visual]) -> bool:
        if not XRenderFindVisualFormat:
            return False

        xrender_format = XRenderFindVisualFormat(self.canvas.display._display, visual)
        return xrender_format and xrender_format.contents.direct.alphaMask != 0

    def get_visual_info(self) -> glx.XVisualInfo:
        return glx.glXGetVisualFromFBConfig(self.canvas.display._display, self.fbconfig).contents  # noqa: SLF001

    def create_context(self, share: XlibContext | None) -> XlibContext:
        return XlibContext(self, share)

    def compatible(self, canvas: XlibCanvas) -> bool:
        # TODO check more
        return isinstance(canvas, XlibCanvas)

    def _create_glx_context(self, _share: None) -> NoReturn:
        raise NotImplementedError

    def is_complete(self) -> bool:
        return True


class XlibContext(Context):  # noqa: D101
    x_display: Display
    glx_context: glx.GLXContext
    glx_window: glx.GLXWindow | None
    _use_video_sync: bool
    _vsync: bool
    config: XlibDisplayConfig

    _have_SGI_swap_control: bool  # noqa: N815
    _have_EXT_swap_control: bool  # noqa: N815
    _have_MESA_swap_control: bool  # noqa: N815
    _have_SGI_video_sync: bool  # noqa: N815

    def __init__(self, config: XlibDisplayConfig, share: XlibContext | None) -> None:  # noqa: D107
        super().__init__(config, share)

        self.x_display = config.canvas.display._display  # noqa: SLF001

        self.glx_context = self._create_glx_context(share)
        if not self.glx_context:
            # TODO: Check Xlib error generated
            msg = 'Could not create GL context'
            raise gl.ContextException(msg)

        self._have_SGI_video_sync = config.glx_info.have_extension('GLX_SGI_video_sync')
        self._have_SGI_swap_control = config.glx_info.have_extension('GLX_SGI_swap_control')
        self._have_EXT_swap_control = config.glx_info.have_extension('GLX_EXT_swap_control')
        self._have_MESA_swap_control = config.glx_info.have_extension('GLX_MESA_swap_control')

        # In order of preference:
        # 1. GLX_EXT_swap_control (more likely to work where video_sync will not)
        # 2. GLX_MESA_swap_control (same as above, but supported by MESA drivers)
        # 3. GLX_SGI_video_sync (does not work on Intel 945GM, but that has EXT)
        # 4. GLX_SGI_swap_control (cannot be disabled once enabled)
        self._use_video_sync = (self._have_SGI_video_sync and
                                not (self._have_EXT_swap_control or self._have_MESA_swap_control))

        # XXX Mandate that vsync defaults on across all platforms.
        self._vsync = True

        self.glx_window = None

    def is_direct(self) -> bool:
        return bool(glx.glXIsDirect(self.x_display, self.glx_context))

    def _create_glx_context(self, share: XlibContext | None) -> glx.GLXContext:
        if share:
            share_context = share.glx_context
        else:
            share_context = None

        attribs = []
        if self.config.major_version is not None:
            attribs.extend([glxext_arb.GLX_CONTEXT_MAJOR_VERSION_ARB, self.config.major_version])
        if self.config.minor_version is not None:
            attribs.extend([glxext_arb.GLX_CONTEXT_MINOR_VERSION_ARB, self.config.minor_version])

        if self.config.opengl_api == "gl":
            attribs.extend([glxext_arb.GLX_CONTEXT_PROFILE_MASK_ARB, glxext_arb.GLX_CONTEXT_CORE_PROFILE_BIT_ARB])
        elif self.config.opengl_api == "gles":
            attribs.extend([glxext_arb.GLX_CONTEXT_PROFILE_MASK_ARB, glxext_arb.GLX_CONTEXT_ES2_PROFILE_BIT_EXT])

        flags = 0
        if self.config.forward_compatible:
            flags |= glxext_arb.GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB
        if self.config.debug:
            flags |= glxext_arb.GLX_CONTEXT_DEBUG_BIT_ARB

        if flags:
            attribs.extend([glxext_arb.GLX_CONTEXT_FLAGS_ARB, flags])
        attribs.append(0)
        attribs = (c_int * len(attribs))(*attribs)

        return glxext_arb.glXCreateContextAttribsARB(self.config.canvas.display._display,  # noqa: SLF001
                                                     self.config.fbconfig, share_context, True, attribs)

    def attach(self, canvas: XlibCanvas) -> None:
        if canvas is self.canvas:
            return

        super().attach(canvas)

        self.glx_window = glx.glXCreateWindow(self.x_display, self.config.fbconfig, canvas.x_window, None)
        self.set_current()

    def set_current(self) -> None:
        glx.glXMakeContextCurrent(self.x_display, self.glx_window, self.glx_window, self.glx_context)
        super().set_current()

    def detach(self) -> None:
        if not self.canvas:
            return

        self.set_current()
        gl.glFlush()

        super().detach()

        glx.glXMakeContextCurrent(self.x_display, 0, 0, None)
        if self.glx_window:
            glx.glXDestroyWindow(self.x_display, self.glx_window)
            self.glx_window = None

    def destroy(self) -> None:
        super().destroy()
        if self.glx_window:
            glx.glXDestroyWindow(self.x_display, self.glx_window)
            self.glx_window = None
        if self.glx_context:
            glx.glXDestroyContext(self.x_display, self.glx_context)
            self.glx_context = None

    def set_vsync(self, vsync: bool = True) -> None:
        self._vsync = vsync
        interval = vsync and 1 or 0
        try:
            if not self._use_video_sync and self._have_EXT_swap_control:
                glxext_arb.glXSwapIntervalEXT(self.x_display, glx.glXGetCurrentDrawable(), interval)
            elif not self._use_video_sync and self._have_MESA_swap_control:
                glxext_mesa.glXSwapIntervalMESA(interval)
            elif self._have_SGI_swap_control:
                glxext_arb.glXSwapIntervalSGI(interval)
        except lib.MissingFunctionException as e:
            warnings.warn(str(e))

    def get_vsync(self) -> bool:
        return self._vsync

    def _wait_vsync(self) -> None:
        if self._vsync and self._have_SGI_video_sync and self._use_video_sync:
            count = c_uint()
            glxext_arb.glXGetVideoSyncSGI(byref(count))
            glxext_arb.glXWaitVideoSyncSGI(2, (count.value + 1) % 2, byref(count))

    def flip(self) -> None:
        if not self.glx_window:
            return

        if self._vsync:
            self._wait_vsync()

        glx.glXSwapBuffers(self.x_display, self.glx_window)



"""Print details of a media file that pyglet can open (requires FFmpeg).

Usage::

    media_info.py <filename>

"""
import sys
import pyglet


def print_ffmpeg_info():
    from pyglet.media import have_ffmpeg

    if have_ffmpeg():
        from pyglet.media.codecs import ffmpeg
        print(f'Using FFmpeg version {ffmpeg.get_version()}')
    else:
        print('FFmpeg not available')
        print('https://www.ffmpeg.org/download.html\n')


def print_source_info(source):
    if source.info:
        if source.info.title:
            print(f'Title: {source.info.title}')
        if source.info.album:
            print(f'Album: {source.info.album}')
        if source.info.author:
            print(f'Author: {source.info.author}')
        if source.info.year:
            print(f'Year: {source.info.year}')
        if source.info.track:
            print(f'Track: {source.info.track}')
        if source.info.genre:
            print(f'Genre: {source.info.genre}')
        if source.info.copyright:
            print(f'Copyright: {source.info.copyright}')
        if source.info.comment:
            print(f'Comment: {source.info.comment}')

    if source.audio_format:
        af = source.audio_format
        print(f'Audio: {af.channels} channel(s), {af.sample_size} bits, {af.sample_rate:.02f} Hz')

    if source.video_format:
        vf = source.video_format
        if vf.frame_rate:
            frame_rate = f'{vf.frame_rate:.02f}'
        else:
            frame_rate = 'unknown'
        if vf.sample_aspect >= 1:
            display_width = vf.sample_aspect * vf.width
            display_height = vf.height
        else:
            display_width = vf.width
            display_height = vf.sample_aspect / vf.height
        print(f'Video: {vf.width}x{vf.height} at aspect {vf.sample_aspect!r} (displays at {display_width}x{display_height}), {frame_rate} fps')

    hours = int(source.duration / 3600)
    minutes = int(source.duration / 60) % 60
    seconds = int(source.duration) % 60
    milliseconds = int(source.duration * 1000) % 1000
    print(f'Duration: {hours}:{minutes:02d}:{seconds:02d}.{milliseconds:03d}')


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print(__doc__)
        print_ffmpeg_info()
        sys.exit(1)

    print_ffmpeg_info()

    filename = sys.argv[1]
    try:
        source = pyglet.media.load(filename, streaming=True)
        print_source_info(source)
        del source
    except pyglet.media.codecs.MediaException:
        print(f'Codec not available to open: {filename}')



"""Test that multiple windows share objects by default.
"""

from ctypes import *

import pyglet

from pyglet import window
from pyglet.gl import *


def create_context(share):
    display = pyglet.display.get_display()
    screen = display.get_default_screen()
    config = screen.get_best_config()
    return config.create_context(share)


def test_context_share_texture():
    w1 = window.Window(200, 200)
    w1.switch_to()
    textures = c_uint()
    glGenTextures(1, byref(textures))
    texture = textures.value

    glBindTexture(GL_TEXTURE_2D, texture)
    data = (c_ubyte * 4)()
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 1, 1, 0, GL_RGBA, GL_UNSIGNED_BYTE, data)
    assert glIsTexture(texture)

    w2 = window.Window(200, 200)
    w2.switch_to()
    assert glIsTexture(texture)

    glDeleteTextures(1, byref(textures))
    assert not glIsTexture(texture)

    w1.switch_to()
    assert not glIsTexture(texture)

    w1.close()
    w2.close()


def test_context_noshare_texture():
    w1 = window.Window(200, 200)
    w1.switch_to()
    textures = c_uint()
    glGenTextures(1, byref(textures))
    texture = textures.value

    glBindTexture(GL_TEXTURE_2D, texture)
    data = (c_ubyte * 4)()
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 1, 1, 0, GL_RGBA, GL_UNSIGNED_BYTE, data)
    assert glIsTexture(texture)

    w2 = window.Window(200, 200, context=create_context(None))
    w2.switch_to()
    assert not glIsTexture(texture)

    w1.close()
    w2.close()



"""Platform-specific support components.

These consist of:

1. ctypes bindings for datastructures and functions
2. pyglet-specific wrapper functions around raw ctypes calls
3. vendored libraries in original or modified forms

When documenting these modules:

1. Use minimal formatting in any docstrings
2. Leave licenses at the tops of files in place

Simple docstrings with minimal formatting are best because:

1. No web doc is built for pyglet.lib
2. The docstrings will be used to debug complex platform issues
3. IDEs mangle formatting in any hover tooltips while debugging
"""
